# -*- coding: utf-8 -*-
"""
OpenAlgo Technical Indicators Library

A high-performance technical analysis library with NumPy and Numba optimizations.
Provides TradingView-like syntax for easy use.
"""

import numpy as np
import pandas as pd
from typing import Union, Tuple, Optional

# Import all indicator classes
from .trend import (SMA, EMA, WMA, DEMA, TEMA, Supertrend, Ichimoku, HMA, VWMA, 
                   ALMA, KAMA, ZLEMA, T3, FRAMA, ChandeKrollStop, TRIMA, 
                   McGinley, VIDYA, Alligator, MovingAverageEnvelopes)
from .momentum import (RSI, MACD, Stochastic, CCI, WilliamsR, BOP, 
                      ElderRay, Fisher, CRSI)
from .volatility import (ATR, BollingerBands, Keltner, Donchian,
                        Chaikin, NATR, RVI as VolatilityRVI, ULTOSC, TRANGE, MASS,
                        BBPercent, BBWidth, ChandelierExit,
                        HistoricalVolatility, UlcerIndex, STARC)
from .volume import (OBV, OBVSmoothed, VWAP, MFI, ADL, CMF, EMV, FI, NVI, PVI, VOLOSC, VROC,
                    KlingerVolumeOscillator, PriceVolumeTrend, RVOL)
from .oscillators import (CMO, TRIX, UO, AO, AC, PPO, PO, DPO, AROONOSC,
                         StochRSI, RVI, CHO, CHOP, KST, TSI, VI, 
                         GatorOscillator, STC, Coppock)
from .statistics import (LINREG, LRSLOPE, CORREL, BETA, VAR, TSF, MEDIAN, MODE, MedianBands)
from .hybrid import (ADX, Aroon, PivotPoints, SAR, DMI,
                    WilliamsFractals, RWI)
from .utils import (crossover, crossunder, highest, lowest, change, roc, 
                   sma as utils_sma, ema as utils_ema, stdev, validate_input,
                   exrem, flip, valuewhen, rising, falling, cross)


class TechnicalAnalysis:
    """
    Main technical analysis interface providing TradingView-like syntax
    
    Usage:
    ------
    from openalgo import ta
    
    # Trend indicators
    sma_20 = ta.sma(close, 20)
    ema_50 = ta.ema(close, 50)
    [supertrend, direction] = ta.supertrend(high, low, close, 10, 3)
    
    # Momentum indicators
    rsi_14 = ta.rsi(close, 14)
    [macd_line, signal_line, histogram] = ta.macd(close, 12, 26, 9)
    
    # Volatility indicators
    [upper, middle, lower] = ta.bbands(close, 20, 2)
    atr_14 = ta.atr(high, low, close, 14)
    
    # Volume indicators
    obv_values = ta.obv(close, volume)
    vwap_values = ta.vwap(high, low, close, volume)
    """
    
    def __init__(self):
        # Initialize all indicator classes
        # Trend indicators
        self._sma = SMA()
        self._ema = EMA()
        self._wma = WMA()
        self._dema = DEMA()
        self._tema = TEMA()
        self._hma = HMA()
        self._vwma = VWMA()
        self._alma = ALMA()
        self._kama = KAMA()
        self._zlema = ZLEMA()
        self._t3 = T3()
        self._frama = FRAMA()
        self._supertrend = Supertrend()
        self._ichimoku = Ichimoku()
        self._chande_kroll_stop = ChandeKrollStop()
        self._trima = TRIMA()
        self._mcginley_dynamic = McGinley()
        self._vidya = VIDYA()
        self._alligator = Alligator()
        self._ma_envelopes = MovingAverageEnvelopes()
        
        # Momentum indicators
        self._rsi = RSI()
        self._macd = MACD()
        self._stochastic = Stochastic()
        self._cci = CCI()
        self._williams_r = WilliamsR()
        self._balance_of_power = BOP()
        self._elder_ray = ElderRay()
        self._fisher_transform = Fisher()
        self._connors_rsi = CRSI()
        
        # Volatility indicators
        self._atr = ATR()
        self._bbands = BollingerBands()
        self._keltner = Keltner()
        self._donchian = Donchian()
        self._chaikin_volatility = Chaikin()
        self._natr = NATR()
        self._rvi = RVI()  # Oscillator RVI (Relative Vigor Index)
        self._rvol = RVOL()  # Relative Volume
        self._ultosc = ULTOSC()
        self._trange = TRANGE()
        self._mass = MASS()
        self._bbands_percent_b = BBPercent()
        self._bbands_bandwidth = BBWidth()
        self._chandelier_exit = ChandelierExit()
        self._hv = HistoricalVolatility()
        self._ulcer_index = UlcerIndex()
        self._starc_bands = STARC()
        
        # Volume indicators
        self._obv = OBV()
        self._obv_smoothed = OBVSmoothed()
        self._vwap = VWAP()
        self._mfi = MFI()
        self._adl = ADL()
        self._cmf = CMF()
        self._emv = EMV()
        self._fi = FI()
        self._nvi = NVI()
        self._pvi = PVI()
        self._vo = VOLOSC()
        self._vroc = VROC()
        self._klinger_vo = KlingerVolumeOscillator()
        self._pvt = PriceVolumeTrend()
        
        # Oscillators
        self._cmo = CMO()
        self._trix = TRIX()
        self._uo = UO()
        self._ao = AO()
        self._ac = AC()
        self._ppo = PPO()
        self._po = PO()
        self._dpo = DPO()
        self._aroonosc = AROONOSC()
        self._stoch_rsi = StochRSI()
        self._rvi_osc = RVI()
        self._chaikin_osc = CHO()
        self._chop = CHOP()
        self._kst = KST()
        self._tsi = TSI()
        self._vi = VI()
        self._stc = STC()
        self._gator_oscillator = GatorOscillator()
        self._coppock = Coppock()
        
        # Statistical indicators
        self._linearreg = LINREG()
        self._linearreg_slope = LRSLOPE()
        self._correl = CORREL()
        self._beta = BETA()
        self._var = VAR()
        self._tsf = TSF()
        self._median = MEDIAN()
        self._median_bands = MedianBands()
        self._mode = MODE()
        
        # Hybrid indicators
        self._adx = ADX()
        self._aroon = Aroon()
        self._pivot_points = PivotPoints()
        self._sar = SAR()
        self._dmi = DMI()
        self._williams_fractals = WilliamsFractals()
        self._random_walk_index = RWI()
    
    # =================== TREND INDICATORS ===================
    
    def sma(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Simple Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            SMA values in the same format as input
        """
        return self._sma.calculate(data, period)
    
    def ema(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Exponential Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            EMA values in the same format as input
        """
        return self._ema.calculate(data, period)
    
    def wma(self, data: Union[np.ndarray, pd.Series, list], period: int) -> np.ndarray:
        """
        Weighted Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        np.ndarray
            Array of WMA values
        """
        return self._wma.calculate(data, period)
    
    def dema(self, data: Union[np.ndarray, pd.Series, list], period: int) -> np.ndarray:
        """
        Double Exponential Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        np.ndarray
            Array of DEMA values
        """
        return self._dema.calculate(data, period)
    
    def tema(self, data: Union[np.ndarray, pd.Series, list], period: int) -> np.ndarray:
        """
        Triple Exponential Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        np.ndarray
            Array of TEMA values
        """
        return self._tema.calculate(data, period)
    
    def supertrend(self, high: Union[np.ndarray, pd.Series, list],
                   low: Union[np.ndarray, pd.Series, list],
                   close: Union[np.ndarray, pd.Series, list],
                   period: int = 10, multiplier: float = 3.0) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]:
        """
        Supertrend Indicator
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        period : int, default=10
            ATR period
        multiplier : float, default=3.0
            ATR multiplier
            
        Returns:
        --------
        Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]
            (supertrend values, direction values) in the same format as input
        """
        return self._supertrend.calculate(high, low, close, period, multiplier)
    
    def ichimoku(self, high: Union[np.ndarray, pd.Series, list],
                 low: Union[np.ndarray, pd.Series, list],
                 close: Union[np.ndarray, pd.Series, list],
                 conversion_periods: int = 9, base_periods: int = 26,
                 lagging_span2_periods: int = 52, displacement: int = 26) -> Tuple[np.ndarray, ...]:
        """
        Ichimoku Cloud - matches TradingView exactly
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        conversion_periods : int, default=9
            Conversion Line Length (TradingView: conversionPeriods)
        base_periods : int, default=26
            Base Line Length (TradingView: basePeriods)
        lagging_span2_periods : int, default=52
            Leading Span B Length (TradingView: laggingSpan2Periods)
        displacement : int, default=26
            Lagging Span displacement (TradingView: displacement)
            
        Returns:
        --------
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
            (conversion_line, base_line, leading_span_a, leading_span_b, lagging_span)
        """
        return self._ichimoku.calculate(high, low, close, conversion_periods, 
                                       base_periods, lagging_span2_periods, displacement)
    
    def hma(self, data: Union[np.ndarray, pd.Series, list], period: int) -> np.ndarray:
        """Hull Moving Average"""
        return self._hma.calculate(data, period)
    
    def vwma(self, data: Union[np.ndarray, pd.Series, list],
             volume: Union[np.ndarray, pd.Series, list], period: int) -> np.ndarray:
        """Volume Weighted Moving Average"""
        return self._vwma.calculate(data, volume, period)
    
    def alma(self, data: Union[np.ndarray, pd.Series, list], 
             period: int = 21, offset: float = 0.85, sigma: float = 6.0) -> np.ndarray:
        """Arnaud Legoux Moving Average"""
        return self._alma.calculate(data, period, offset, sigma)
    
    def kama(self, data: Union[np.ndarray, pd.Series, list],
             length: int = 14, fast_length: int = 2, slow_length: int = 30) -> np.ndarray:
        """Kaufman's Adaptive Moving Average - matches TradingView exactly"""
        return self._kama.calculate(data, length, fast_length, slow_length)
    
    def zlema(self, data: Union[np.ndarray, pd.Series, list], period: int) -> np.ndarray:
        """Zero Lag Exponential Moving Average"""
        return self._zlema.calculate(data, period)
    
    def t3(self, data: Union[np.ndarray, pd.Series, list],
           period: int = 21, v_factor: float = 0.7) -> np.ndarray:
        """T3 Moving Average"""
        return self._t3.calculate(data, period, v_factor)
    
    def frama(self, high: Union[np.ndarray, pd.Series, list],
              low: Union[np.ndarray, pd.Series, list], period: int = 26) -> Union[np.ndarray, pd.Series]:
        """Fractal Adaptive Moving Average - matches TradingView exactly"""
        return self._frama.calculate(high, low, period)
    
    # =================== MOMENTUM INDICATORS ===================
    
    def rsi(self, data: Union[np.ndarray, pd.Series, list], period: int = 14) -> np.ndarray:
        """
        Relative Strength Index
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int, default=14
            Number of periods for RSI calculation
            
        Returns:
        --------
        np.ndarray
            Array of RSI values
        """
        return self._rsi.calculate(data, period)
    
    def macd(self, data: Union[np.ndarray, pd.Series, list], 
             fast_period: int = 12, slow_period: int = 26, 
             signal_period: int = 9) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Moving Average Convergence Divergence
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        fast_period : int, default=12
            Period for fast EMA
        slow_period : int, default=26
            Period for slow EMA
        signal_period : int, default=9
            Period for signal line EMA
            
        Returns:
        --------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            (macd_line, signal_line, histogram)
        """
        return self._macd.calculate(data, fast_period, slow_period, signal_period)
    
    def stochastic(self, high: Union[np.ndarray, pd.Series, list],
                   low: Union[np.ndarray, pd.Series, list],
                   close: Union[np.ndarray, pd.Series, list],
                   k_period: int = 14, d_period: int = 3) -> Tuple[np.ndarray, np.ndarray]:
        """
        Stochastic Oscillator
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        k_period : int, default=14
            Period for %K calculation
        d_period : int, default=3
            Period for %D calculation
            
        Returns:
        --------
        Tuple[np.ndarray, np.ndarray]
            (k_percent, d_percent)
        """
        return self._stochastic.calculate(high, low, close, k_period, d_period)
    
    def cci(self, high: Union[np.ndarray, pd.Series, list],
            low: Union[np.ndarray, pd.Series, list],
            close: Union[np.ndarray, pd.Series, list],
            period: int = 20) -> np.ndarray:
        """
        Commodity Channel Index
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        period : int, default=20
            Number of periods for CCI calculation
            
        Returns:
        --------
        np.ndarray
            Array of CCI values
        """
        return self._cci.calculate(high, low, close, period)
    
    def williams_r(self, high: Union[np.ndarray, pd.Series, list],
                   low: Union[np.ndarray, pd.Series, list],
                   close: Union[np.ndarray, pd.Series, list],
                   period: int = 14) -> np.ndarray:
        """
        Williams %R
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        period : int, default=14
            Number of periods for Williams %R calculation
            
        Returns:
        --------
        np.ndarray
            Array of Williams %R values
        """
        return self._williams_r.calculate(high, low, close, period)
    
    # =================== VOLATILITY INDICATORS ===================
    
    def atr(self, high: Union[np.ndarray, pd.Series, list],
            low: Union[np.ndarray, pd.Series, list],
            close: Union[np.ndarray, pd.Series, list],
            period: int = 14) -> np.ndarray:
        """
        Average True Range
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        period : int, default=14
            Number of periods for ATR calculation
            
        Returns:
        --------
        np.ndarray
            Array of ATR values
        """
        return self._atr.calculate(high, low, close, period)
    
    def bbands(self, data: Union[np.ndarray, pd.Series, list],
               period: int = 20, std_dev: float = 2.0) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Bollinger Bands
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int, default=20
            Number of periods for moving average and standard deviation
        std_dev : float, default=2.0
            Number of standard deviations for the bands
            
        Returns:
        --------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            (upper_band, middle_band, lower_band)
        """
        return self._bbands.calculate(data, period, std_dev)
    
    def keltner(self, high: Union[np.ndarray, pd.Series, list],
                        low: Union[np.ndarray, pd.Series, list],
                        close: Union[np.ndarray, pd.Series, list],
                        ema_period: int = 20, atr_period: int = 10, 
                        multiplier: float = 2.0) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Keltner Channel
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        ema_period : int, default=20
            Period for the EMA calculation
        atr_period : int, default=10
            Period for the ATR calculation
        multiplier : float, default=2.0
            Multiplier for the ATR
            
        Returns:
        --------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            (upper_channel, middle_line, lower_channel)
        """
        return self._keltner.calculate(high, low, close, ema_period, atr_period, multiplier)
    
    def donchian(self, high: Union[np.ndarray, pd.Series, list],
                         low: Union[np.ndarray, pd.Series, list],
                         period: int = 20) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Donchian Channel
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        period : int, default=20
            Number of periods for the channel calculation
            
        Returns:
        --------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            (upper_channel, middle_line, lower_channel)
        """
        return self._donchian.calculate(high, low, period)
    
    def chaikin(self, high: Union[np.ndarray, pd.Series, list],
                          low: Union[np.ndarray, pd.Series, list],
                          ema_period: int = 10, roc_period: int = 10) -> np.ndarray:
        """Chaikin Volatility"""
        return self._chaikin_volatility.calculate(high, low, ema_period, roc_period)
    
    def natr(self, high: Union[np.ndarray, pd.Series, list],
             low: Union[np.ndarray, pd.Series, list],
             close: Union[np.ndarray, pd.Series, list],
             period: int = 14) -> np.ndarray:
        """Normalized Average True Range"""
        return self._natr.calculate(high, low, close, period)
    
    def rvol(self, volume: Union[np.ndarray, pd.Series, list], period: int = 20) -> Union[np.ndarray, pd.Series]:
        """Relative Volume"""
        return self._rvol.calculate(volume, period)
    
    def ultimate_oscillator(self, high: Union[np.ndarray, pd.Series, list],
                           low: Union[np.ndarray, pd.Series, list],
                           close: Union[np.ndarray, pd.Series, list],
                           period1: int = 7, period2: int = 14, period3: int = 28) -> np.ndarray:
        """Ultimate Oscillator"""
        return self._ultosc.calculate(high, low, close, period1, period2, period3)
    
    
    def true_range(self, high: Union[np.ndarray, pd.Series, list],
                   low: Union[np.ndarray, pd.Series, list],
                   close: Union[np.ndarray, pd.Series, list]) -> np.ndarray:
        """True Range"""
        return self._trange.calculate(high, low, close)
    
    def massindex(self, high: Union[np.ndarray, pd.Series, list],
                   low: Union[np.ndarray, pd.Series, list],
                   length: int = 10) -> np.ndarray:
        """Mass Index (Pine Script v6 formula)"""
        return self._mass.calculate(high, low, length)
    
    # =================== VOLUME INDICATORS ===================
    
    def obv(self, close: Union[np.ndarray, pd.Series, list],
            volume: Union[np.ndarray, pd.Series, list]) -> np.ndarray:
        """
        On Balance Volume
        
        Parameters:
        -----------
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        volume : Union[np.ndarray, pd.Series, list]
            Volume data
            
        Returns:
        --------
        np.ndarray
            Array of OBV values
        """
        return self._obv.calculate(close, volume)
    
    def obv_smoothed(self, close: Union[np.ndarray, pd.Series, list],
                     volume: Union[np.ndarray, pd.Series, list],
                     ma_type: str = "None",
                     ma_length: int = 20,
                     bb_length: int = 20,
                     bb_mult: float = 2.0) -> Union[Tuple[np.ndarray, np.ndarray, np.ndarray], np.ndarray]:
        """
        On Balance Volume with Smoothing Options (TradingView Pine Script Implementation)
        
        Parameters:
        -----------
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        volume : Union[np.ndarray, pd.Series, list]
            Volume data
        ma_type : str, default "None"
            Moving average type: "None", "SMA", "SMA + Bollinger Bands", "EMA", "SMMA (RMA)", "WMA", "VWMA"
        ma_length : int, default 20
            Moving average length
        bb_length : int, default 20
            Bollinger Bands length (only used for "SMA + Bollinger Bands")
        bb_mult : float, default 2.0
            Bollinger Bands multiplier (only used for "SMA + Bollinger Bands")
            
        Returns:
        --------
        Union[Tuple[np.ndarray, np.ndarray, np.ndarray], np.ndarray]
            If ma_type is "SMA + Bollinger Bands": (obv_smoothed, bb_upper, bb_lower)
            Otherwise: obv_smoothed
        """
        return self._obv_smoothed.calculate(close, volume, ma_type, ma_length, bb_length, bb_mult)
    
    def vwap(self, high: Union[np.ndarray, pd.Series, list],
             low: Union[np.ndarray, pd.Series, list],
             close: Union[np.ndarray, pd.Series, list],
             volume: Union[np.ndarray, pd.Series, list],
             anchor: str = "Session",
             source: str = "hlc3",
             stdev_mult_1: float = 1.0,
             stdev_mult_2: float = 2.0,
             stdev_mult_3: float = 3.0,
             percent_mult_1: float = 0.236,
             percent_mult_2: float = 0.382,
             percent_mult_3: float = 0.618) -> np.ndarray:
        """
        Volume Weighted Average Price - TradingView Pine Script v6 Implementation
        
        Advanced VWAP with session-based anchoring, standard deviation and percentage bands.
        Matches TradingView Pine Script v6 functionality exactly.
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        volume : Union[np.ndarray, pd.Series, list]
            Volume data
        anchor : str, default="Session"
            Session anchoring type: "Session", "Week", "Month", "Quarter", "Year", 
            "12M", "6M", "3M", "D", "4H", "1H", "30m", "15m", "5m", "1m"
        source : str, default="hlc3"
            Price source: "hlc3", "hl2", "ohlc4", "close"
        stdev_mult_1 : float, default=1.0
            Standard deviation multiplier for first band
        stdev_mult_2 : float, default=2.0
            Standard deviation multiplier for second band
        stdev_mult_3 : float, default=3.0
            Standard deviation multiplier for third band
        percent_mult_1 : float, default=0.236
            Percentage multiplier for first percentage band
        percent_mult_2 : float, default=0.382
            Percentage multiplier for second percentage band
        percent_mult_3 : float, default=0.618
            Percentage multiplier for third percentage band
            
        Returns:
        --------
        np.ndarray
            Array of VWAP values
            
        Notes:
        ------
        TradingView Pine Script v6 Features:
        - Session-based anchoring for automatic reset points
        - Multiple price sources (hlc3, hl2, ohlc4, close)
        - Standard deviation bands for volatility analysis
        - Percentage bands for level analysis
        - Volume validation and error handling
        
        Use calculate_with_bands() method for band calculations:
        vwap_result = ta.vwap.calculate_with_bands(high, low, close, volume, ...)
        """
        return self._vwap.calculate(high, low, close, volume, source, anchor)
    
    def mfi(self, high: Union[np.ndarray, pd.Series, list],
            low: Union[np.ndarray, pd.Series, list],
            close: Union[np.ndarray, pd.Series, list],
            volume: Union[np.ndarray, pd.Series, list],
            period: int = 14) -> np.ndarray:
        """
        Money Flow Index
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        volume : Union[np.ndarray, pd.Series, list]
            Volume data
        period : int, default=14
            Number of periods for MFI calculation
            
        Returns:
        --------
        np.ndarray
            Array of MFI values
        """
        return self._mfi.calculate(high, low, close, volume, period)
    
    def adl(self, high: Union[np.ndarray, pd.Series, list],
            low: Union[np.ndarray, pd.Series, list],
            close: Union[np.ndarray, pd.Series, list],
            volume: Union[np.ndarray, pd.Series, list]) -> np.ndarray:
        """Accumulation/Distribution Line"""
        return self._adl.calculate(high, low, close, volume)
    
    def cmf(self, high: Union[np.ndarray, pd.Series, list],
            low: Union[np.ndarray, pd.Series, list],
            close: Union[np.ndarray, pd.Series, list],
            volume: Union[np.ndarray, pd.Series, list],
            period: int = 20) -> np.ndarray:
        """Chaikin Money Flow"""
        return self._cmf.calculate(high, low, close, volume, period)
    
    def emv(self, high: Union[np.ndarray, pd.Series, list],
            low: Union[np.ndarray, pd.Series, list],
            volume: Union[np.ndarray, pd.Series, list],
            length: int = 14, divisor: int = 10000) -> Union[np.ndarray, pd.Series]:
        """Ease of Movement - matches TradingView exactly"""
        return self._emv.calculate(high, low, volume, length, divisor)
    
    def force_index(self, close: Union[np.ndarray, pd.Series, list],
                    volume: Union[np.ndarray, pd.Series, list],
                    length: int = 13) -> Union[np.ndarray, pd.Series]:
        """Elder Force Index - matches TradingView exactly"""
        return self._fi.calculate(close, volume, length)
    
    def nvi(self, close: Union[np.ndarray, pd.Series, list],
            volume: Union[np.ndarray, pd.Series, list]) -> np.ndarray:
        """Negative Volume Index (Pine Script Implementation)"""
        return self._nvi.calculate(close, volume)
    
    def nvi_with_ema(self, close: Union[np.ndarray, pd.Series, list],
                    volume: Union[np.ndarray, pd.Series, list],
                    ema_length: int = 255) -> Tuple[np.ndarray, np.ndarray]:
        """Negative Volume Index with EMA (Complete Pine Script Implementation)"""
        return self._nvi.calculate_with_ema(close, volume, ema_length)
    
    def pvi(self, close: Union[np.ndarray, pd.Series, list],
            volume: Union[np.ndarray, pd.Series, list],
            initial_value: float = 100.0) -> np.ndarray:
        """Positive Volume Index (TradingView Pine Script Implementation)"""
        return self._pvi.calculate(close, volume, initial_value)
    
    def pvi_with_signal(self, close: Union[np.ndarray, pd.Series, list],
                       volume: Union[np.ndarray, pd.Series, list],
                       initial_value: float = 100.0,
                       signal_type: str = "EMA",
                       signal_length: int = 255) -> Tuple[np.ndarray, np.ndarray]:
        """Positive Volume Index with Signal Line (Complete TradingView Pine Script Implementation)"""
        return self._pvi.calculate_with_signal(close, volume, initial_value, signal_type, signal_length)
    
    def volosc(self, volume: Union[np.ndarray, pd.Series, list],
                         short_length: int = 5, long_length: int = 10,
                         check_volume_validity: bool = True) -> np.ndarray:
        """
        Volume Oscillator - TradingView Pine Script v6 Implementation
        
        Parameters:
        -----------
        volume : Union[np.ndarray, pd.Series, list]
            Volume data
        short_length : int, default=5
            Short EMA length (TradingView default)
        long_length : int, default=10
            Long EMA length (TradingView default)
        check_volume_validity : bool, default=True
            Check for valid volume data (TradingView style validation)
            
        Returns:
        --------
        np.ndarray
            Volume Oscillator values
            
        Notes:
        ------
        TradingView Pine Script v6 Formula:
        short = ta.ema(volume, shortlen)
        long = ta.ema(volume, longlen)
        osc = 100 * (short - long) / long
        """
        return self._vo.calculate(volume, short_length, long_length, check_volume_validity)
    
    def vroc(self, volume: Union[np.ndarray, pd.Series, list], period: int = 25) -> np.ndarray:
        """Volume Rate of Change"""
        return self._vroc.calculate(volume, period)
    
    def kvo(self, high: Union[np.ndarray, pd.Series, list],
            low: Union[np.ndarray, pd.Series, list],
            close: Union[np.ndarray, pd.Series, list],
            volume: Union[np.ndarray, pd.Series, list],
            trig_len: int = 13, fast_x: int = 34, slow_x: int = 55) -> Tuple[np.ndarray, np.ndarray]:
        """
        Klinger Volume Oscillator - matches TradingView exactly
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        volume : Union[np.ndarray, pd.Series, list]
            Volume data
        trig_len : int, default=13
            Trigger line EMA period (TradingView: TrigLen)
        fast_x : int, default=34
            Fast EMA period (TradingView: FastX)
        slow_x : int, default=55
            Slow EMA period (TradingView: SlowX)
            
        Returns:
        --------
        Tuple[np.ndarray, np.ndarray]
            (kvo, trigger) values
        """
        return self._klinger_vo.calculate(high, low, close, volume, trig_len, fast_x, slow_x)
    
    # =================== OSCILLATORS ===================
    
    def cmo(self, data: Union[np.ndarray, pd.Series, list], period: int = 14) -> np.ndarray:
        """Chande Momentum Oscillator"""
        return self._cmo.calculate(data, period)
    
    def trix(self, data: Union[np.ndarray, pd.Series, list], length: int = 18) -> np.ndarray:
        """TRIX (TradingView Pine Script v6) - triple EMA rate of change of log prices * 10000"""
        return self._trix.calculate(data, length)
    
    def uo_oscillator(self, high: Union[np.ndarray, pd.Series, list],
                     low: Union[np.ndarray, pd.Series, list],
                     close: Union[np.ndarray, pd.Series, list],
                     period1: int = 7, period2: int = 14, period3: int = 28) -> np.ndarray:
        """Ultimate Oscillator"""
        return self._uo.calculate(high, low, close, period1, period2, period3)
    
    def awesome_oscillator(self, high: Union[np.ndarray, pd.Series, list],
                          low: Union[np.ndarray, pd.Series, list],
                          fast_period: int = 5, slow_period: int = 34) -> np.ndarray:
        """Awesome Oscillator"""
        return self._ao.calculate(high, low, fast_period, slow_period)
    
    def accelerator_oscillator(self, high: Union[np.ndarray, pd.Series, list],
                              low: Union[np.ndarray, pd.Series, list],
                              period: int = 5) -> np.ndarray:
        """Accelerator Oscillator"""
        return self._ac.calculate(high, low, period)
    
    def ppo(self, data: Union[np.ndarray, pd.Series, list],
            fast_period: int = 12, slow_period: int = 26,
            signal_period: int = 9) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Percentage Price Oscillator"""
        return self._ppo.calculate(data, fast_period, slow_period, signal_period)
    
    def po(self, data: Union[np.ndarray, pd.Series, list],
                        fast_period: int = 10, slow_period: int = 20,
                        ma_type: str = "SMA") -> np.ndarray:
        """Price Oscillator"""
        return self._po.calculate(data, fast_period, slow_period, ma_type)
    
    def dpo(self, data: Union[np.ndarray, pd.Series, list], 
           period: int = 21, is_centered: bool = False) -> Union[np.ndarray, pd.Series]:
        """Detrended Price Oscillator - matches TradingView exactly"""
        return self._dpo.calculate(data, period, is_centered)
    
    def aroon_oscillator(self, high: Union[np.ndarray, pd.Series, list],
                        low: Union[np.ndarray, pd.Series, list],
                        period: int = 14) -> np.ndarray:
        """Aroon Oscillator"""
        return self._aroonosc.calculate(high, low, period)
    
    # =================== STATISTICAL INDICATORS ===================
    
    def linreg(self, data: Union[np.ndarray, pd.Series, list], period: int = 14) -> np.ndarray:
        """Linear Regression"""
        return self._linearreg.calculate(data, period)
    
    def lrslope(self, data: Union[np.ndarray, pd.Series, list], 
               period: int = 100, interval: int = 1) -> np.ndarray:
        """Linear Regression Slope - matches TradingView exactly"""
        return self._linearreg_slope.calculate(data, period, interval)
    
    def correlation(self, data1: Union[np.ndarray, pd.Series, list],
                   data2: Union[np.ndarray, pd.Series, list],
                   period: int = 20) -> np.ndarray:
        """Pearson Correlation Coefficient"""
        return self._correl.calculate(data1, data2, period)
    
    def beta(self, asset: Union[np.ndarray, pd.Series, list],
             market: Union[np.ndarray, pd.Series, list],
             period: int = 252) -> np.ndarray:
        """Beta Coefficient"""
        return self._beta.calculate(asset, market, period)
    
    def variance(self, data: Union[np.ndarray, pd.Series, list], 
                lookback: int = 20, mode: str = "PR", ema_period: int = 20,
                filter_lookback: int = 20, ema_length: int = 14,
                return_components: bool = False) -> Union[np.ndarray, tuple]:
        """
        Variance - TradingView Pine Script v4 Implementation
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (close prices)
        lookback : int, default=20
            Variance lookback period
        mode : str, default="PR"
            Variance mode: "LR" for Logarithmic Returns, "PR" for Price
        ema_period : int, default=20
            EMA period for variance smoothing
        filter_lookback : int, default=20
            Lookback period for variance filter (z-score calculation)  
        ema_length : int, default=14
            EMA length for z-score smoothing
        return_components : bool, default=False
            If True, returns (variance, ema_variance, zscore, ema_zscore, stdev)
            
        Returns:
        --------
        Union[np.ndarray, tuple]
            Variance values or tuple of all components
        """
        return self._var.calculate(data, lookback, mode, ema_period, filter_lookback, ema_length, return_components)
    
    def tsf(self, data: Union[np.ndarray, pd.Series, list], period: int = 14) -> np.ndarray:
        """Time Series Forecast"""
        return self._tsf.calculate(data, period)
    
    def median(self, data: Union[np.ndarray, pd.Series, list], period: int = 3) -> np.ndarray:
        """Rolling Median (Pine Script v6)"""
        return self._median.calculate(data, period)
    
    def median_bands(self, high: Union[np.ndarray, pd.Series, list],
                    low: Union[np.ndarray, pd.Series, list],
                    close: Union[np.ndarray, pd.Series, list],
                    source: Optional[Union[np.ndarray, pd.Series, list]] = None,
                    median_length: int = 3,
                    atr_length: int = 14,
                    atr_mult: float = 2.0) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Median with Bands and EMA (Pine Script v6 Complete)"""
        return self._median_bands.calculate_with_bands(high, low, close, source, 
                                                       median_length, atr_length, atr_mult)
    
    def mode(self, data: Union[np.ndarray, pd.Series, list], 
             period: int = 20, bins: int = 10) -> np.ndarray:
        """Rolling Mode"""
        return self._mode.calculate(data, period, bins)
    
    # =================== HYBRID INDICATORS ===================
    
    def adx(self, high: Union[np.ndarray, pd.Series, list],
                   low: Union[np.ndarray, pd.Series, list],
                   close: Union[np.ndarray, pd.Series, list],
                   period: int = 14) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Average Directional Index System (+DI, -DI, ADX)"""
        return self._adx.calculate(high, low, close, period)
    
    def aroon(self, high: Union[np.ndarray, pd.Series, list],
                     low: Union[np.ndarray, pd.Series, list],
                     period: int = 25) -> Tuple[np.ndarray, np.ndarray]:
        """Aroon Indicator (Up, Down)"""
        return self._aroon.calculate(high, low, period)
    
    def pivot_points(self, high: Union[np.ndarray, pd.Series, list],
                     low: Union[np.ndarray, pd.Series, list],
                     close: Union[np.ndarray, pd.Series, list]) -> Tuple[np.ndarray, ...]:
        """Pivot Points (Pivot, R1, S1, R2, S2, R3, S3)"""
        return self._pivot_points.calculate(high, low, close)
    
    def dmi(self, high: Union[np.ndarray, pd.Series, list],
                            low: Union[np.ndarray, pd.Series, list],
                            close: Union[np.ndarray, pd.Series, list],
                            period: int = 14) -> Tuple[np.ndarray, np.ndarray]:
        """Directional Movement Index (+DI, -DI)"""
        return self._dmi.calculate(high, low, close, period)
    
    def psar(self, high: Union[np.ndarray, pd.Series, list],
             low: Union[np.ndarray, pd.Series, list],
             acceleration: float = 0.02, maximum: float = 0.2) -> np.ndarray:
        """Parabolic SAR (values only)"""
        sar_values, trend = self._sar.calculate(high, low, acceleration, maximum)
        return sar_values
    
    def ckstop(self, high: Union[np.ndarray, pd.Series, list],
              low: Union[np.ndarray, pd.Series, list],
              close: Union[np.ndarray, pd.Series, list],
              p: int = 10, x: float = 1.0, q: int = 9) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]:
        """Chande Kroll Stop - matches TradingView exactly"""
        return self._chande_kroll_stop.calculate(high, low, close, p, x, q)
    
    # =================== NEW MOMENTUM INDICATORS ===================
    
    def bop(self, open_prices: Union[np.ndarray, pd.Series, list],
           high: Union[np.ndarray, pd.Series, list],
           low: Union[np.ndarray, pd.Series, list],
           close: Union[np.ndarray, pd.Series, list]) -> Union[np.ndarray, pd.Series]:
        """Balance of Power"""
        return self._balance_of_power.calculate(open_prices, high, low, close)
    
    def elderray(self, high: Union[np.ndarray, pd.Series, list],
                low: Union[np.ndarray, pd.Series, list],
                close: Union[np.ndarray, pd.Series, list],
                period: int = 13) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]:
        """Elder Ray Index (Bull/Bear Power)"""
        return self._elder_ray.calculate(high, low, close, period)
    
    def fisher(self, high: Union[np.ndarray, pd.Series, list],
              low: Union[np.ndarray, pd.Series, list],
              length: int = 9) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]:
        """Fisher Transform - matches TradingView exactly"""
        return self._fisher_transform.calculate(high, low, length)
    
    def crsi(self, data: Union[np.ndarray, pd.Series, list],
            lenrsi: int = 3, lenupdown: int = 2, 
            lenroc: int = 100) -> Union[np.ndarray, pd.Series]:
        """Connors RSI - matches TradingView exactly"""
        return self._connors_rsi.calculate(data, lenrsi, lenupdown, lenroc)
    
    # =================== NEW VOLUME INDICATORS ===================
    
    def pvt(self, close: Union[np.ndarray, pd.Series, list],
           volume: Union[np.ndarray, pd.Series, list]) -> Union[np.ndarray, pd.Series]:
        """
        Price Volume Trend (TradingView Pine Script Implementation)
        
        Parameters:
        -----------
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        volume : Union[np.ndarray, pd.Series, list]
            Volume data
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            PVT values - Formula: vt = ta.cum(ta.change(src)/src[1]*volume)
        """
        return self._pvt.calculate(close, volume)
    
    # =================== NEW OSCILLATORS ===================
    
    def stochrsi(self, data: Union[np.ndarray, pd.Series, list], 
                rsi_period: int = 14, stoch_period: int = 14,
                k_period: int = 3, d_period: int = 3) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]:
        """Stochastic RSI"""
        return self._stoch_rsi.calculate(data, rsi_period, stoch_period, k_period, d_period)
    
    def rvi(self, open_prices: Union[np.ndarray, pd.Series, list],
           high: Union[np.ndarray, pd.Series, list],
           low: Union[np.ndarray, pd.Series, list],
           close: Union[np.ndarray, pd.Series, list],
           period: int = 10) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]:
        """
        Relative Vigor Index (TradingView Pine Script Implementation)
        
        Parameters:
        -----------
        open_prices : Union[np.ndarray, pd.Series, list]
            Opening prices
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        period : int, default=10
            Period for RVI calculation
            
        Returns:
        --------
        Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]
            (rvi, signal) - Formula: rvi = math.sum(ta.swma(close-open), len) / math.sum(ta.swma(high-low), len)
        """
        return self._rvi_osc.calculate(open_prices, high, low, close, period)
    
    def cho(self, high: Union[np.ndarray, pd.Series, list],
           low: Union[np.ndarray, pd.Series, list],
           close: Union[np.ndarray, pd.Series, list],
           volume: Union[np.ndarray, pd.Series, list],
           fast_period: int = 3, slow_period: int = 10) -> Union[np.ndarray, pd.Series]:
        """Chaikin Oscillator"""
        return self._chaikin_osc.calculate(high, low, close, volume, fast_period, slow_period)
    
    # =================== MISSING FUNCTIONS ALIASES ===================
    
    # Create aliases for the renamed functions
    def bbwidth(self, data, period=20, std_dev=2.0):
        """Bollinger Bands Bandwidth (alias for bbands_bandwidth)"""
        return self._bbands_bandwidth.calculate(data, period, std_dev) if hasattr(self, '_bbands_bandwidth') else None
    
    def bbpercent(self, data, period=20, std_dev=2.0):
        """Bollinger Bands %B (alias for bbands_percent_b)"""
        return self._bbands_percent_b.calculate(data, period, std_dev) if hasattr(self, '_bbands_percent_b') else None
    
    def mcginley(self, data, period=14):
        """McGinley Dynamic (alias for mcginley_dynamic)"""
        return self._mcginley_dynamic.calculate(data, period) if hasattr(self, '_mcginley_dynamic') else None
        
    def starc(self, high, low, close, ma_period=5, atr_period=15, multiplier=1.33):
        """STARC Bands (TradingView Pine Script v2) - uses SMA(5) + ATR(15) * 1.33"""
        return self._starc_bands.calculate(high, low, close, ma_period, atr_period, multiplier) if hasattr(self, '_starc_bands') else None
    
    def ulcerindex(self, data, length=14, smooth_length=14, signal_length=52, signal_type="SMA", return_signal=False):
        """Ulcer Index (TradingView Pine Script v4) - measures downside risk with optional signal line"""
        return self._ulcer_index.calculate(data, length, smooth_length, signal_length, signal_type, return_signal) if hasattr(self, '_ulcer_index') else None
        
    def rwi(self, high, low, close, period=14):
        """Random Walk Index (alias for random_walk_index)"""
        return self._random_walk_index.calculate(high, low, close, period) if hasattr(self, '_random_walk_index') else None
        
    def fractals(self, high, low, periods=2):
        """Williams Fractals (alias for williams_fractals) - matches TradingView exactly"""  
        return self._williams_fractals.calculate(high, low, periods) if hasattr(self, '_williams_fractals') else None
        
    def trima(self, data, period=20):
        """Triangular Moving Average (alias)"""
        return self._trima.calculate(data, period) if hasattr(self, '_trima') else None
        
    def vidya(self, data, period=14, alpha=0.2):
        """Variable Index Dynamic Average (alias)"""
        return self._vidya.calculate(data, period, alpha) if hasattr(self, '_vidya') else None
        
    def alligator(self, data, jaw_period=13, jaw_shift=8, teeth_period=8, teeth_shift=5, lips_period=5, lips_shift=3):
        """Bill Williams Alligator (alias)"""
        return self._alligator.calculate(data, jaw_period, jaw_shift, teeth_period, teeth_shift, lips_period, lips_shift) if hasattr(self, '_alligator') else None
        
    def ma_envelopes(self, data, period=20, percentage=2.5, ma_type='SMA'):
        """Moving Average Envelopes (alias)"""
        return self._ma_envelopes.calculate(data, period, percentage, ma_type) if hasattr(self, '_ma_envelopes') else None
        
    def chandelier_exit(self, high, low, close, period=22, multiplier=3.0):
        """Chandelier Exit (alias)"""
        return self._chandelier_exit.calculate(high, low, close, period, multiplier) if hasattr(self, '_chandelier_exit') else None
        
    def hv(self, close, length=10, annual=365, per=1):
        """Historical Volatility (alias) - matches TradingView exactly"""
        return self._hv.calculate(close, length, annual, per) if hasattr(self, '_hv') else None
        
    def chop(self, high, low, close, period=14):
        """Choppiness Index (alias)"""
        return self._chop.calculate(high, low, close, period) if hasattr(self, '_chop') else None
        
    def kst(self, data, roclen1=10, roclen2=15, roclen3=20, roclen4=30, smalen1=10, smalen2=10, smalen3=10, smalen4=15, siglen=9):
        """Know Sure Thing - matches TradingView exactly"""
        return self._kst.calculate(data, roclen1, roclen2, roclen3, roclen4, smalen1, smalen2, smalen3, smalen4, siglen) if hasattr(self, '_kst') else None
        
    def tsi(self, data, long_period=25, short_period=13, signal_period=13):
        """True Strength Index (alias)"""
        return self._tsi.calculate(data, long_period, short_period, signal_period) if hasattr(self, '_tsi') else None
        
    def vi(self, high: Union[np.ndarray, pd.Series, list],
               low: Union[np.ndarray, pd.Series, list],
               close: Union[np.ndarray, pd.Series, list],
               period: int = 14) -> Union[Tuple[np.ndarray, np.ndarray], None]:
        """
        Vortex Indicator (VI+ and VI-) - TradingView Pine Script v6 Implementation
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        period : int, default=14
            Period for VI calculation (TradingView default)
            
        Returns:
        --------
        Union[Tuple[np.ndarray, np.ndarray], None]
            (vi_plus, vi_minus) or None if not available
            
        Notes:
        ------
        TradingView Pine Script v6 Formula:
        VMP = math.sum( math.abs( high - low[1]), period_ )
        VMM = math.sum( math.abs( low - high[1]), period_ )
        STR = math.sum( ta.atr(1), period_ )
        VIP = VMP / STR
        VIM = VMM / STR
        """
        return self._vi.calculate(high, low, close, period) if hasattr(self, '_vi') else None
        
    def stc(self, data, fast_length=23, slow_length=50, cycle_length=10, d1_length=3, d2_length=3):
        """Schaff Trend Cycle (TradingView Pine Script v4) - cyclical oscillator combining MACD and stochastics"""
        return self._stc.calculate(data, fast_length, slow_length, cycle_length, d1_length, d2_length) if hasattr(self, '_stc') else None
        
    def gator_oscillator(self, high, low, jaw_period=13, teeth_period=8, lips_period=5):
        """Gator Oscillator (alias) - matches TradingView exactly"""
        return self._gator_oscillator.calculate(high, low, jaw_period, teeth_period, lips_period) if hasattr(self, '_gator_oscillator') else None
        
    
    # =================== UTILITY FUNCTIONS ===================
    
    def crossover(self, series1: Union[np.ndarray, pd.Series, list], 
                  series2: Union[np.ndarray, pd.Series, list]) -> np.ndarray:
        """
        Check if series1 crosses over series2
        
        Parameters:
        -----------
        series1 : Union[np.ndarray, pd.Series, list]
            First series
        series2 : Union[np.ndarray, pd.Series, list]
            Second series
            
        Returns:
        --------
        np.ndarray
            Boolean array indicating crossover points
        """
        series1 = validate_input(series1)
        series2 = validate_input(series2)
        return crossover(series1, series2)
    
    def crossunder(self, series1: Union[np.ndarray, pd.Series, list], 
                   series2: Union[np.ndarray, pd.Series, list]) -> np.ndarray:
        """
        Check if series1 crosses under series2
        
        Parameters:
        -----------
        series1 : Union[np.ndarray, pd.Series, list]
            First series
        series2 : Union[np.ndarray, pd.Series, list]
            Second series
            
        Returns:
        --------
        np.ndarray
            Boolean array indicating crossunder points
        """
        series1 = validate_input(series1)
        series2 = validate_input(series2)
        return crossunder(series1, series2)
    
    def highest(self, data: Union[np.ndarray, pd.Series, list], period: int) -> np.ndarray:
        """
        Highest value over a period
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Input data
        period : int
            Window size
            
        Returns:
        --------
        np.ndarray
            Array of highest values
        """
        data = validate_input(data)
        return highest(data, period)
    
    def lowest(self, data: Union[np.ndarray, pd.Series, list], period: int) -> np.ndarray:
        """
        Lowest value over a period
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Input data
        period : int
            Window size
            
        Returns:
        --------
        np.ndarray
            Array of lowest values
        """
        data = validate_input(data)
        return lowest(data, period)
    
    def change(self, data: Union[np.ndarray, pd.Series, list], length: int = 1) -> np.ndarray:
        """
        Change in value over a specified number of periods
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Input data
        length : int, default=1
            Number of periods to look back
            
        Returns:
        --------
        np.ndarray
            Array of change values
        """
        data = validate_input(data)
        return change(data, length)
    
    def roc(self, data: Union[np.ndarray, pd.Series, list], length: int) -> np.ndarray:
        """
        Rate of Change (ROC)
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Input data
        length : int
            Number of periods to look back
            
        Returns:
        --------
        np.ndarray
            Array of ROC values as percentages
        """
        data = validate_input(data)
        return roc(data, length)
    
    def stdev(self, data: Union[np.ndarray, pd.Series, list], period: int) -> np.ndarray:
        """
        Rolling standard deviation
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Input data
        period : int
            Window size for standard deviation calculation
            
        Returns:
        --------
        np.ndarray
            Array of standard deviation values
        """
        data = validate_input(data)
        return stdev(data, period)
    
    def exrem(self, primary: Union[np.ndarray, pd.Series, list], 
              secondary: Union[np.ndarray, pd.Series, list]) -> np.ndarray:
        """
        Excess Removal function - eliminates excessive signals
        
        Parameters:
        -----------
        primary : Union[np.ndarray, pd.Series, list]
            Primary signal array (boolean-like)
        secondary : Union[np.ndarray, pd.Series, list]
            Secondary signal array (boolean-like)
            
        Returns:
        --------
        np.ndarray
            Boolean array with excess signals removed
        """
        primary = validate_input(primary)
        secondary = validate_input(secondary)
        return exrem(primary, secondary)
    
    def flip(self, primary: Union[np.ndarray, pd.Series, list], 
             secondary: Union[np.ndarray, pd.Series, list]) -> np.ndarray:
        """
        Flip function - creates a toggle state based on two signals
        
        Parameters:
        -----------
        primary : Union[np.ndarray, pd.Series, list]
            Primary signal array (boolean-like)
        secondary : Union[np.ndarray, pd.Series, list]
            Secondary signal array (boolean-like)
            
        Returns:
        --------
        np.ndarray
            Boolean array representing flip state
        """
        primary = validate_input(primary)
        secondary = validate_input(secondary)
        return flip(primary, secondary)
    
    def valuewhen(self, expr: Union[np.ndarray, pd.Series, list], 
                  array: Union[np.ndarray, pd.Series, list], n: int = 1) -> np.ndarray:
        """
        Returns the value of array when expr was true for the nth most recent time
        
        Parameters:
        -----------
        expr : Union[np.ndarray, pd.Series, list]
            Expression array (boolean-like)
        array : Union[np.ndarray, pd.Series, list]
            Value array to sample from
        n : int, default=1
            Which occurrence to get (1 = most recent, 2 = second most recent, etc.)
            
        Returns:
        --------
        np.ndarray
            Array of values when condition was true
        """
        expr = validate_input(expr)
        array = validate_input(array)
        return valuewhen(expr, array, n)
    
    def rising(self, data: Union[np.ndarray, pd.Series, list], length: int) -> np.ndarray:
        """
        Check if data is rising (current value > value n periods ago)
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Input data series
        length : int
            Number of periods to look back
            
        Returns:
        --------
        np.ndarray
            Boolean array indicating rising periods
        """
        data = validate_input(data)
        return rising(data, length)
    
    def falling(self, data: Union[np.ndarray, pd.Series, list], length: int) -> np.ndarray:
        """
        Check if data is falling (current value < value n periods ago)
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Input data series
        length : int
            Number of periods to look back
            
        Returns:
        --------
        np.ndarray
            Boolean array indicating falling periods
        """
        data = validate_input(data)
        return falling(data, length)
    
    def cross(self, series1: Union[np.ndarray, pd.Series, list], 
              series2: Union[np.ndarray, pd.Series, list]) -> np.ndarray:
        """
        Check if series1 crosses series2 (either direction)
        Combines crossover and crossunder functionality
        
        Parameters:
        -----------
        series1 : Union[np.ndarray, pd.Series, list]
            First series
        series2 : Union[np.ndarray, pd.Series, list]
            Second series
            
        Returns:
        --------
        np.ndarray
            Boolean array indicating cross points (both over and under)
        """
        series1 = validate_input(series1)
        series2 = validate_input(series2)
        return cross(series1, series2)
    
    def coppock(self, data: Union[np.ndarray, pd.Series, list], 
                wma_length: int = 10, long_roc_length: int = 14, 
                short_roc_length: int = 11) -> Union[np.ndarray, pd.Series]:
        """
        Coppock Curve - Long-term momentum indicator
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        wma_length : int, default=10
            WMA Length for final smoothing
        long_roc_length : int, default=14
            Long RoC Length  
        short_roc_length : int, default=11
            Short RoC Length
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            Coppock Curve values in the same format as input
        """
        return self._coppock.calculate(data, wma_length, long_roc_length, short_roc_length)


# Create global instance for easy access
ta = TechnicalAnalysis()

# Make indicator classes available for advanced users
__all__ = [
    'ta', 'TechnicalAnalysis',
    # Trend indicators
    'SMA', 'EMA', 'WMA', 'DEMA', 'TEMA', 'HMA', 'VWMA', 'ALMA', 'KAMA', 'ZLEMA', 'T3', 'FRAMA',
    'Supertrend', 'Ichimoku', 'ChandeKrollStop', 'TRIMA', 'McGinley', 'VIDYA', 'Alligator', 'MovingAverageEnvelopes',
    # Momentum indicators  
    'RSI', 'MACD', 'Stochastic', 'CCI', 'WilliamsR', 'BOP', 'ElderRay', 
    'Fisher', 'CRSI',
    # Volatility indicators
    'ATR', 'BollingerBands', 'Keltner', 'Donchian', 'Chaikin', 'NATR', 
    'RVI', 'ULTOSC', 'TRANGE', 'MASS', 'BBPercent', 'BBWidth', 'STARC',
    'ChandelierExit', 'HistoricalVolatility', 'UlcerIndex',
    # Volume indicators
    'OBV', 'VWAP', 'MFI', 'ADL', 'CMF', 'EMV', 'FI', 'NVI', 'PVI', 'VOLOSC', 'VROC',
    'KlingerVolumeOscillator', 'PriceVolumeTrend',
    # Oscillators
    'ROC', 'CMO', 'TRIX', 'UO', 'AO', 'AC', 'PPO', 'PO', 'DPO', 'AROONOSC',
    'StochRSI', 'RVI', 'CHO', 'CHOP', 'KST', 'TSI', 'VI', 'STC', 'GatorOscillator', 'Coppock',
    # Statistical indicators
    'LINREG', 'LRSLOPE', 'CORREL', 'BETA', 'VAR', 'TSF', 'MEDIAN', 'MODE',
    # Hybrid indicators
    'ADX', 'Aroon', 'PivotPoints', 'SAR', 'DMI', 'WilliamsFractals', 'RWI',
    # Utility functions
    'crossover', 'crossunder', 'highest', 'lowest', 'change', 'roc', 'stdev',
    'exrem', 'flip', 'valuewhen', 'rising', 'falling', 'cross'
]


# -*- coding: utf-8 -*-
"""
OpenAlgo WebSocket API Documentation - Feed Methods
    https://docs.openalgo.in
"""

import json
import threading
import time
from typing import List, Dict, Any, Callable, Optional
import websocket
from .base import BaseAPI

class FeedAPI(BaseAPI):
    """
    Market data feed API methods for OpenAlgo using WebSockets.
    Inherits from the BaseAPI class.
    """

    def __init__(self, api_key, host="http://127.0.0.1:5000", version="v1", ws_port=8765, ws_url=None, verbose=False):
        """
        Initialize the FeedAPI object with API key and optionally a host URL, API version, and WebSocket details.

        Attributes:
        - api_key (str): User's API key.
        - host (str): Base URL for the API endpoints. Defaults to localhost.
        - version (str): API version. Defaults to "v1".
        - ws_port (int): WebSocket server port. Defaults to 8765.
        - ws_url (str, optional): Custom WebSocket URL. If provided, this overrides host and ws_port settings.
        - verbose (int): Logging verbosity level. Defaults to False.
            - 0 or False: Silent mode (errors only)
            - 1 or True: Basic info (connection, auth, subscription status)
            - 2: Full debug (all market data updates)
        """
        super().__init__(api_key, host, version)

        # Verbosity control
        self.verbose = int(verbose) if verbose is not False else 0
        
        # WebSocket configuration
        self.ws_port = ws_port
        
        # Use custom WebSocket URL if provided
        if ws_url:
            self.ws_url = ws_url
        else:
            # Extract host without protocol for WebSocket
            if host.startswith("http://"):
                self.ws_host = host[7:]
            elif host.startswith("https://"):
                self.ws_host = host[8:]
            else:
                self.ws_host = host
                
            # Remove any path component and port if present
            self.ws_host = self.ws_host.split('/')[0].split(':')[0]
            
            # Create default WebSocket URL
            self.ws_url = f"ws://{self.ws_host}:{self.ws_port}"
        self.ws = None
        self.connected = False
        self.authenticated = False
        self.ws_thread = None
        
        # Message management
        self.message_queue = []
        self.lock = threading.Lock()
        
        # Data storage
        self.ltp_data = {}  # Structure: {'EXCHANGE:SYMBOL': {'price': price, 'timestamp': timestamp}}
        self.quotes_data = {}  # Structure: {'EXCHANGE:SYMBOL': {'open': open, 'high': high, 'low': low, 'close': close, 'ltp': ltp, 'timestamp': timestamp}}
        self.depth_data = {}  # Structure: {'EXCHANGE:SYMBOL': {'ltp': ltp, 'timestamp': timestamp, 'depth': {'buy': [...], 'sell': [...]}}}
        
        # Callback registry
        self.ltp_callback = None
        self.quote_callback = None
        self.quotes_callback = None
        self.depth_callback = None

    def _log(self, level: int, category: str, message: str) -> None:
        """
        Internal logging method with verbosity control.

        Args:
            level (int): Required verbosity level (1=basic, 2=debug)
            category (str): Log category for alignment (WS, AUTH, SUB, LTP, QUOTE, DEPTH, ERROR)
            message (str): The message to log
        """
        if self.verbose >= level:
            # Fixed width category for alignment
            cat_width = 6
            formatted_cat = f"[{category}]".ljust(cat_width + 2)
            print(f"{formatted_cat} {message}")

    def connect(self) -> bool:
        """
        Connect to the WebSocket server and authenticate.
        
        Returns:
            bool: True if connection and authentication are successful, False otherwise.
        """
        try:
            def on_message(ws, message):
                self._process_message(message)
                
            def on_error(ws, error):
                self._log(1, "ERROR", f"WebSocket error: {error}")

            def on_open(ws):
                self._log(1, "WS", f"Connected to {self.ws_url}")
                self.connected = True
                self._authenticate()

            def on_close(ws, close_status_code, close_reason):
                self._log(1, "WS", f"Disconnected from {self.ws_url}")
                self.connected = False
                self.authenticated = False
            
            # Initialize WebSocket connection
            self.ws = websocket.WebSocketApp(
                self.ws_url,
                on_message=on_message,
                on_error=on_error,
                on_open=on_open,
                on_close=on_close
            )
            
            # Start WebSocket connection in a separate thread
            self.ws_thread = threading.Thread(target=self.ws.run_forever)
            self.ws_thread.daemon = True
            self.ws_thread.start()
            
            # Wait for connection to establish
            timeout = 5
            start_time = time.time()
            while not self.connected and time.time() - start_time < timeout:
                time.sleep(0.1)
            
            if not self.connected:
                self._log(1, "ERROR", "Failed to connect to WebSocket server")
                return False
                
            # Wait for authentication to complete
            timeout = 5
            start_time = time.time()
            while not self.authenticated and time.time() - start_time < timeout and self.connected:
                time.sleep(0.1)
                
            return self.authenticated
            
        except Exception as e:
            self._log(1, "ERROR", f"Error connecting to WebSocket: {e}")
            return False

    def disconnect(self) -> None:
        """Disconnect from the WebSocket server."""
        if self.ws:
            self.ws.close()
            # Wait for websocket to close
            timeout = 2
            start_time = time.time()
            while self.connected and time.time() - start_time < timeout:
                time.sleep(0.1)
            self.ws = None
            self.connected = False
            self.authenticated = False

    def _authenticate(self) -> None:
        """Authenticate with the WebSocket server using the API key."""
        if not self.connected:
            self._log(1, "ERROR", "Cannot authenticate: not connected")
            return

        auth_msg = {
            "action": "authenticate",
            "api_key": self.api_key
        }

        self._log(1, "AUTH", f"Authenticating with API key: {self.api_key[:8]}...{self.api_key[-8:]}")
        self.ws.send(json.dumps(auth_msg))

    def _process_message(self, message_str: str) -> None:
        """
        Process incoming WebSocket messages.
        
        Args:
            message_str (str): The message string received from the WebSocket.
        """
        try:
            message = json.loads(message_str)
            
            # Handle authentication response
            if message.get("type") == "auth":
                if message.get("status") == "success":
                    self.authenticated = True
                    broker = message.get("broker", "unknown")
                    user_id = message.get("user_id", "unknown")
                    self._log(1, "AUTH", f"Success | Broker: {broker} | User: {user_id}")
                    self._log(2, "AUTH", f"Full response: {message}")
                else:
                    self._log(1, "ERROR", f"Authentication failed: {message.get('message', 'Unknown error')}")
                return

            # Handle subscription response
            if message.get("type") == "subscribe":
                subs = message.get("subscriptions", [])
                for sub in subs:
                    sym = sub.get("symbol", "?")
                    exch = sub.get("exchange", "?")
                    status = sub.get("status", "?")
                    mode = sub.get("mode", 0)
                    mode_name = {1: "LTP", 2: "Quote", 3: "Depth"}.get(mode, "Unknown")
                    self._log(1, "SUB", f"{exch}:{sym} | Mode: {mode_name} | Status: {status}")
                self._log(2, "SUB", f"Full response: {message}")
                return
                
            # Handle market data
            if message.get("type") == "market_data":
                exchange = message.get("exchange")
                symbol = message.get("symbol")
                if exchange and symbol:
                    mode = message.get("mode")
                    market_data = message.get("data", {})
                    
                    # Handle LTP data (mode 1)
                    if mode == 1 and "ltp" in market_data:
                        with self.lock:
                            # Get LTP and timestamp from the message
                            ltp = market_data.get("ltp")
                            timestamp = market_data.get("timestamp", int(time.time() * 1000))
                            
                            # Store both price and timestamp with format 'EXCHANGE:SYMBOL'
                            symbol_key = f"{exchange}:{symbol}"
                            self.ltp_data[symbol_key] = {
                                'price': ltp,
                                'timestamp': timestamp
                            }

                            self._log(2, "LTP", f"{symbol_key:<20} | LTP: {ltp}")
                        
                        # Invoke callback if set
                        if self.ltp_callback:
                            try:
                                # Create a clean market data update without redundant fields
                                clean_data = {
                                    'type': 'market_data',
                                    'symbol': symbol,
                                    'exchange': exchange,
                                    'mode': mode,
                                    'data': {
                                        'ltp': ltp,
                                        'timestamp': timestamp
                                    }
                                }
                                # Include LTT if available in original data
                                if 'ltt' in market_data:
                                    clean_data['data']['ltt'] = market_data['ltt']
                                # Check if ltt is in the nested data structure (which seems to be the case)
                                elif 'ltt' in market_data.get('data', {}):
                                    clean_data['data']['ltt'] = market_data['data']['ltt']
                                    
                                # Pass the cleaned message to callback
                                self.ltp_callback(clean_data)
                            except Exception as e:
                                self._log(1, "ERROR", f"LTP callback error: {str(e)}")                 
                    # Handle Quotes data (mode 2)
                    elif mode == 2:
                        with self.lock:
                            # Extract quote data fields
                            quote_data = {
                                'open': market_data.get("open", 0),
                                'high': market_data.get("high", 0),
                                'low': market_data.get("low", 0),
                                'close': market_data.get("close", 0),
                                'ltp': market_data.get("ltp", 0),
                                'volume': market_data.get("volume", 0),
                                'timestamp': market_data.get("timestamp", int(time.time() * 1000))
                            }
                            
                            # Store quote data with format 'EXCHANGE:SYMBOL'
                            symbol_key = f"{exchange}:{symbol}"
                            self.quotes_data[symbol_key] = quote_data

                            self._log(2, "QUOTE", f"{symbol_key:<20} | O: {quote_data['open']:<10} H: {quote_data['high']:<10} L: {quote_data['low']:<10} C: {quote_data['close']:<10} LTP: {quote_data['ltp']}")
                        
                        # Invoke callback if set
                        if self.quote_callback:
                            try:
                                # Create a clean market data update without redundant fields
                                clean_data = {
                                    'type': 'market_data',
                                    'symbol': symbol,
                                    'exchange': exchange,
                                    'mode': mode,
                                    'data': quote_data.copy()
                                }
                                # Pass the cleaned message to callback
                                self.quote_callback(clean_data)
                            except Exception as e:
                                self._log(1, "ERROR", f"Quote callback error: {str(e)}")                 
                    # Handle Market Depth data (mode 3)
                    elif mode == 3 and "depth" in market_data:
                        with self.lock:
                            # Extract depth data
                            depth_data = {
                                'ltp': market_data.get("ltp", 0),
                                'timestamp': market_data.get("timestamp", int(time.time() * 1000)),
                                'depth': market_data.get("depth", {"buy": [], "sell": []})
                            }
                            
                            # Store depth data with format 'EXCHANGE:SYMBOL'
                            symbol_key = f"{exchange}:{symbol}"
                            self.depth_data[symbol_key] = depth_data

                            # Log depth data
                            buy_depth = depth_data.get('depth', {}).get('buy', [])
                            sell_depth = depth_data.get('depth', {}).get('sell', [])

                            self._log(2, "DEPTH", f"{symbol_key:<20} | LTP: {depth_data.get('ltp')}")

                            if self.verbose >= 2:
                                # Print buy depth summary
                                print(f"         {'BUY':<30} | {'SELL':<30}")
                                print(f"         {'Price':<10} {'Qty':<10} {'Orders':<8} | {'Price':<10} {'Qty':<10} {'Orders':<8}")
                                print("         " + "-" * 62)
                                max_levels = max(len(buy_depth), len(sell_depth), 1)
                                for i in range(max_levels):
                                    buy_lvl = buy_depth[i] if i < len(buy_depth) else {}
                                    sell_lvl = sell_depth[i] if i < len(sell_depth) else {}
                                    bp = buy_lvl.get('price', '-')
                                    bq = buy_lvl.get('quantity', '-')
                                    bo = buy_lvl.get('orders', '-')
                                    sp = sell_lvl.get('price', '-')
                                    sq = sell_lvl.get('quantity', '-')
                                    so = sell_lvl.get('orders', '-')
                                    print(f"         {str(bp):<10} {str(bq):<10} {str(bo):<8} | {str(sp):<10} {str(sq):<10} {str(so):<8}")
                        
                        # Invoke callback if set
                        if self.depth_callback:
                            try:
                                # Create a clean market data update
                                clean_data = {
                                    'type': 'market_data',
                                    'symbol': symbol,
                                    'exchange': exchange,
                                    'mode': mode,
                                    'data': depth_data.copy()
                                }
                                # Pass the cleaned message to callback
                                self.depth_callback(clean_data)
                            except Exception as e:
                                self._log(1, "ERROR", f"Depth callback error: {str(e)}")
                        
        except json.JSONDecodeError:
            self._log(1, "ERROR", f"Invalid JSON message: {message_str[:100]}...")
        except Exception as e:
            self._log(1, "ERROR", f"Error handling message: {e}")

    def subscribe_ltp(self, instruments: List[Dict[str, Any]], on_data_received: Optional[Callable] = None) -> bool:
        """
        Subscribe to LTP updates for instruments.
        
        Args:
            instruments: List of instrument dictionaries with keys:
                - exchange (str): Exchange code (e.g., 'NSE', 'BSE', 'NFO')
                - symbol (str): Trading symbol
                - exchange_token (str, optional): Exchange token for the instrument
            on_data_received: Callback function for data updates
                
        Returns:
            bool: True if subscription successful, False otherwise
        """
        if not self.connected:
            self._log(1, "ERROR", "Not connected to WebSocket server")
            return False

        if not self.authenticated:
            self._log(1, "ERROR", "Not authenticated with WebSocket server")
            return False

        # Set callback if provided
        if on_data_received:
            self.ltp_callback = on_data_received

        # Subscribe to each instrument individually (matching the working test implementation)
        for instrument in instruments:
            exchange = instrument.get("exchange")
            symbol = instrument.get("symbol")
            exchange_token = instrument.get("exchange_token")

            # If only exchange_token is provided, we need to map it to a symbol
            if not symbol and exchange_token:
                symbol = exchange_token

            if not exchange or not symbol:
                self._log(1, "ERROR", f"Invalid instrument: {instrument}")
                continue

            # Use the exact same message format as the working test
            subscription_msg = {
                "action": "subscribe",
                "symbol": symbol,
                "exchange": exchange,
                "mode": 1,  # 1 for LTP
                "depth": 5  # Default depth level
            }

            self._log(1, "SUB", f"Subscribing {exchange}:{symbol} LTP...")
            try:
                self.ws.send(json.dumps(subscription_msg))

                # Small delay to ensure the message is processed separately (just like the test)
                time.sleep(0.1)
            except Exception as e:
                self._log(1, "ERROR", f"Error subscribing to {exchange}:{symbol}: {e}")
                return False

        return True

    def unsubscribe_ltp(self, instruments: List[Dict[str, Any]]) -> bool:
        """
        Unsubscribe from LTP updates for instruments.

        Args:
            instruments: List of instrument dictionaries with keys:
                - exchange (str): Exchange code (e.g., 'NSE', 'BSE', 'NFO')
                - symbol (str): Trading symbol
                - exchange_token (str, optional): Exchange token for the instrument

        Returns:
            bool: True if unsubscription successful, False otherwise
        """
        if not self.connected or not self.authenticated:
            return False

        # Unsubscribe from each instrument individually (matching the working test implementation)
        for instrument in instruments:
            exchange = instrument.get("exchange")
            symbol = instrument.get("symbol")
            exchange_token = instrument.get("exchange_token")

            # If only exchange_token is provided, we need to map it to a symbol
            if not symbol and exchange_token:
                symbol = exchange_token

            if not exchange or not symbol:
                self._log(1, "ERROR", f"Invalid instrument: {instrument}")
                continue

            self._log(1, "UNSUB", f"Unsubscribing {exchange}:{symbol} LTP")

            # Use the exact same message format as the working test
            unsubscribe_msg = {
                "action": "unsubscribe",
                "symbol": symbol,
                "exchange": exchange,
                "mode": 1  # 1 for LTP
            }

            try:
                self.ws.send(json.dumps(unsubscribe_msg))

                # Clean up the data
                with self.lock:
                    symbol_key = f"{exchange}:{symbol}"
                    if symbol_key in self.ltp_data:
                        del self.ltp_data[symbol_key]

                # Small delay to ensure the message is processed separately
                time.sleep(0.1)
            except Exception as e:
                self._log(1, "ERROR", f"Error unsubscribing {exchange}:{symbol}: {e}")
                return False

        return True
        
    def subscribe_quote(self, instruments: List[Dict[str, Any]], on_data_received: Optional[Callable] = None) -> bool:
        """
        Subscribe to Quote updates for instruments.

        Args:
            instruments: List of instrument dictionaries with keys:
                - exchange (str): Exchange code (e.g., 'NSE', 'BSE', 'NFO')
                - symbol (str): Trading symbol
                - exchange_token (str, optional): Exchange token for the instrument
            on_data_received: Callback function for data updates

        Returns:
            bool: True if subscription request sent successfully
        """
        if not self.connected:
            self._log(1, "ERROR", "Not connected to WebSocket server")
            return False

        if not self.authenticated:
            self._log(1, "ERROR", "Not authenticated with WebSocket server")
            return False

        # Set callback if provided
        if on_data_received:
            self.quote_callback = on_data_received

        # Subscribe to each instrument individually
        for instrument in instruments:
            exchange = instrument.get("exchange")
            symbol = instrument.get("symbol")
            exchange_token = instrument.get("exchange_token")

            # If only exchange_token is provided, we need to map it to a symbol
            if not symbol and exchange_token:
                symbol = exchange_token

            if not exchange or not symbol:
                self._log(1, "ERROR", f"Invalid instrument: {instrument}")
                continue

            # Use the same message format as for LTP but with mode 2 for Quote
            subscription_msg = {
                "action": "subscribe",
                "symbol": symbol,
                "exchange": exchange,
                "mode": 2,  # 2 for Quote
                "depth": 5  # Default depth level
            }

            self._log(1, "SUB", f"Subscribing {exchange}:{symbol} Quote...")
            try:
                self.ws.send(json.dumps(subscription_msg))

                # Small delay to ensure the message is processed separately
                time.sleep(0.1)
            except Exception as e:
                self._log(1, "ERROR", f"Error subscribing to {exchange}:{symbol}: {e}")
                return False

        return True
    
    def unsubscribe_quote(self, instruments: List[Dict[str, Any]]) -> bool:
        """
        Unsubscribe from Quote updates for instruments.

        Args:
            instruments: List of instrument dictionaries with keys:
                - exchange (str): Exchange code (e.g., 'NSE', 'BSE', 'NFO')
                - symbol (str): Trading symbol
                - exchange_token (str, optional): Exchange token for the instrument

        Returns:
            bool: True if unsubscription successful, False otherwise
        """
        if not self.connected or not self.authenticated:
            return False

        # Unsubscribe from each instrument individually
        for instrument in instruments:
            exchange = instrument.get("exchange")
            symbol = instrument.get("symbol")
            exchange_token = instrument.get("exchange_token")

            # If only exchange_token is provided, we need to map it to a symbol
            if not symbol and exchange_token:
                symbol = exchange_token

            if not exchange or not symbol:
                self._log(1, "ERROR", f"Invalid instrument: {instrument}")
                continue

            self._log(1, "UNSUB", f"Unsubscribing {exchange}:{symbol} Quote")

            # Use the same message format as for LTP but with mode 2 for Quote
            unsubscribe_msg = {
                "action": "unsubscribe",
                "symbol": symbol,
                "exchange": exchange,
                "mode": 2  # 2 for Quote
            }

            try:
                self.ws.send(json.dumps(unsubscribe_msg))

                # Clean up the data
                with self.lock:
                    symbol_key = f"{exchange}:{symbol}"
                    if symbol_key in self.quotes_data:
                        del self.quotes_data[symbol_key]

                # Small delay to ensure the message is processed separately
                time.sleep(0.1)
            except Exception as e:
                self._log(1, "ERROR", f"Error unsubscribing {exchange}:{symbol}: {e}")
                return False

        return True
        
    def subscribe_depth(self, instruments: List[Dict[str, Any]], on_data_received: Optional[Callable] = None) -> bool:
        """
        Subscribe to Market Depth updates for instruments.

        Args:
            instruments: List of instrument dictionaries with keys:
                - exchange (str): Exchange code (e.g., 'NSE', 'BSE', 'NFO')
                - symbol (str): Trading symbol
                - exchange_token (str, optional): Exchange token for the instrument
            on_data_received: Callback function for data updates

        Returns:
            bool: True if subscription request sent successfully
        """
        if not self.connected:
            self._log(1, "ERROR", "Not connected to WebSocket server")
            return False

        if not self.authenticated:
            self._log(1, "ERROR", "Not authenticated with WebSocket server")
            return False

        # Set callback if provided
        if on_data_received:
            self.depth_callback = on_data_received

        # Subscribe to each instrument individually
        for instrument in instruments:
            exchange = instrument.get("exchange")
            symbol = instrument.get("symbol")
            exchange_token = instrument.get("exchange_token")

            # If only exchange_token is provided, we need to map it to a symbol
            if not symbol and exchange_token:
                symbol = exchange_token

            if not exchange or not symbol:
                self._log(1, "ERROR", f"Invalid instrument: {instrument}")
                continue

            # Use the same message format as for Quote but with mode 3 for Market Depth
            subscription_msg = {
                "action": "subscribe",
                "symbol": symbol,
                "exchange": exchange,
                "mode": 3,  # 3 for Market Depth
                "depth": 5  # Default depth level
            }

            self._log(1, "SUB", f"Subscribing {exchange}:{symbol} Depth...")
            try:
                self.ws.send(json.dumps(subscription_msg))

                # Small delay to ensure the message is processed separately
                time.sleep(0.1)
            except Exception as e:
                self._log(1, "ERROR", f"Error subscribing to {exchange}:{symbol}: {e}")
                return False

        return True
    
    def unsubscribe_depth(self, instruments: List[Dict[str, Any]]) -> bool:
        """
        Unsubscribe from Market Depth updates for instruments.

        Args:
            instruments: List of instrument dictionaries with keys:
                - exchange (str): Exchange code (e.g., 'NSE', 'BSE', 'NFO')
                - symbol (str): Trading symbol
                - exchange_token (str, optional): Exchange token for the instrument

        Returns:
            bool: True if unsubscription successful, False otherwise
        """
        if not self.connected or not self.authenticated:
            return False

        # Unsubscribe from each instrument individually
        for instrument in instruments:
            exchange = instrument.get("exchange")
            symbol = instrument.get("symbol")
            exchange_token = instrument.get("exchange_token")

            # If only exchange_token is provided, we need to map it to a symbol
            if not symbol and exchange_token:
                symbol = exchange_token

            if not exchange or not symbol:
                self._log(1, "ERROR", f"Invalid instrument: {instrument}")
                continue

            self._log(1, "UNSUB", f"Unsubscribing {exchange}:{symbol} Depth")

            # Use the same message format as for Quote but with mode 3 for Market Depth
            unsubscribe_msg = {
                "action": "unsubscribe",
                "symbol": symbol,
                "exchange": exchange,
                "mode": 3  # 3 for Market Depth
            }

            try:
                self.ws.send(json.dumps(unsubscribe_msg))

                # Clean up the data
                with self.lock:
                    symbol_key = f"{exchange}:{symbol}"
                    if symbol_key in self.depth_data:
                        del self.depth_data[symbol_key]

                # Small delay to ensure the message is processed separately
                time.sleep(0.1)
            except Exception as e:
                self._log(1, "ERROR", f"Error unsubscribing {exchange}:{symbol}: {e}")
                return False

        return True

    def get_ltp(self, exchange: str = None, symbol: str = None) -> Dict[str, Any]:
        """
        Get the latest LTP data in nested format.
        
        Args:
            exchange (str, optional): Filter by exchange
            symbol (str, optional): Filter by symbol (requires exchange to be specified)
            
        Returns:
            dict: Dictionary with LTP data in nested format:
                {"ltp": {"EXCHANGE": {"SYMBOL": {"timestamp": timestamp, "ltp": price}}}}
        """
        with self.lock:
            # Create nested format response
            result = {"ltp": {}}
            
            # Process each item in the data structure
            for symbol_key, data in self.ltp_data.items():
                # Extract exchange and symbol from the key (format: "EXCHANGE:SYMBOL")
                if ":" in symbol_key:
                    parts = symbol_key.split(":")
                    ex = parts[0]  # Exchange
                    sym = parts[1]  # Symbol
                    
                    # Filter by exchange if specified
                    if exchange and ex != exchange:
                        continue
                        
                    # Filter by symbol if specified
                    if symbol and sym != symbol:
                        continue
                    
                    # Initialize exchange dict if not exists
                    if ex not in result["ltp"]:
                        result["ltp"][ex] = {}
                    
                    # Add data to the nested structure
                    result["ltp"][ex][sym] = {
                        "timestamp": data['timestamp'],
                        "ltp": data['price']
                    }
            
            return result
            
    def get_quotes(self, exchange: str = None, symbol: str = None) -> Dict[str, Any]:
        """
        Get the latest Quote data in nested format.
        
        Args:
            exchange (str, optional): Filter by exchange
            symbol (str, optional): Filter by symbol (requires exchange to be specified)
            
        Returns:
            dict: Dictionary with Quote data in nested format:
                {"quote": {"EXCHANGE": {"SYMBOL": {
                    "timestamp": timestamp,
                    "open": open,
                    "high": high,
                    "low": low,
                    "close": close,
                    "ltp": ltp,
                    "volume": volume
                }}}}
        """
        with self.lock:
            # Create nested format response
            result = {"quote": {}}
            
            # Process each item in the data structure
            for symbol_key, data in self.quotes_data.items():
                # Extract exchange and symbol from the key (format: "EXCHANGE:SYMBOL")
                if ":" in symbol_key:
                    parts = symbol_key.split(":")
                    ex = parts[0]  # Exchange
                    sym = parts[1]  # Symbol
                    
                    # Filter by exchange if specified
                    if exchange and ex != exchange:
                        continue
                        
                    # Filter by symbol if specified
                    if symbol and sym != symbol:
                        continue
                    
                    # Initialize exchange dict if not exists
                    if ex not in result["quote"]:
                        result["quote"][ex] = {}
                    
                    # Add data to the nested structure
                    result["quote"][ex][sym] = {
                        "timestamp": data['timestamp'],
                        "open": data['open'],
                        "high": data['high'],
                        "low": data['low'],
                        "close": data['close'],
                        "ltp": data['ltp'],
                        "volume": data.get('volume', 0)
                    }
            
            return result
            
    def get_depth(self, exchange: str = None, symbol: str = None) -> Dict[str, Any]:
        """
        Get the latest Market Depth data in nested format.
        
        Args:
            exchange (str, optional): Filter by exchange
            symbol (str, optional): Filter by symbol (requires exchange to be specified)
            
        Returns:
            dict: Dictionary with Market Depth data in nested format:
                {"depth": {"EXCHANGE": {"SYMBOL": {
                    "timestamp": timestamp,
                    "ltp": ltp,
                    "buyBook": {
                        "1": {"price": price, "qty": quantity, "orders": orders},
                        # Additional levels...
                    },
                    "sellBook": {
                        "1": {"price": price, "qty": quantity, "orders": orders},
                        # Additional levels...
                    }
                }}}}
        """
        with self.lock:
            # Create nested format response
            result = {"depth": {}}
            
            # Process each item in the data structure
            for symbol_key, data in self.depth_data.items():
                # Extract exchange and symbol from the key (format: "EXCHANGE:SYMBOL")
                if ":" in symbol_key:
                    parts = symbol_key.split(":")
                    ex = parts[0]  # Exchange
                    sym = parts[1]  # Symbol
                    
                    # Filter by exchange if specified
                    if exchange and ex != exchange:
                        continue
                        
                    # Filter by symbol if specified
                    if symbol and sym != symbol:
                        continue
                    
                    # Initialize exchange dict if not exists
                    if ex not in result["depth"]:
                        result["depth"][ex] = {}
                    
                    # Initialize the symbol structure
                    result["depth"][ex][sym] = {
                        "timestamp": data.get('timestamp', int(time.time() * 1000)),
                        "ltp": data.get('ltp', 0),
                        "buyBook": {},
                        "sellBook": {}
                    }
                    
                    # Process buy depth book
                    buy_depth = data.get('depth', {}).get('buy', [])
                    for i, level in enumerate(buy_depth):
                        level_num = str(i + 1)
                        result["depth"][ex][sym]["buyBook"][level_num] = {
                            "price": float(level.get('price') or 0),
                            "qty": int(level.get('quantity') or 0),
                            "orders": int(level.get('orders') or 0)
                        }

                    # If there are fewer than 5 levels, add empty levels to complete the structure
                    for i in range(len(buy_depth), 5):
                        level_num = str(i + 1)
                        result["depth"][ex][sym]["buyBook"][level_num] = {
                            "price": 0.0,
                            "qty": 0,
                            "orders": 0
                        }

                    # Process sell depth book
                    sell_depth = data.get('depth', {}).get('sell', [])
                    for i, level in enumerate(sell_depth):
                        level_num = str(i + 1)
                        result["depth"][ex][sym]["sellBook"][level_num] = {
                            "price": float(level.get('price') or 0),
                            "qty": int(level.get('quantity') or 0),
                            "orders": int(level.get('orders') or 0)
                        }

                    # If there are fewer than 5 levels, add empty levels to complete the structure
                    for i in range(len(sell_depth), 5):
                        level_num = str(i + 1)
                        result["depth"][ex][sym]["sellBook"][level_num] = {
                            "price": 0.0,
                            "qty": 0,
                            "orders": 0
                        }
            
            return result



# -*- coding: utf-8 -*-
"""
OpenAlgo REST API Documentation - Base API Class
    https://docs.openalgo.in
"""

import httpx

class BaseAPI:
    """
    Base class to handle all the API calls to OpenAlgo.
    """

    def __init__(self, api_key, host="http://127.0.0.1:5000", version="v1", timeout=120.0):
        """
        Initialize the api object with an API key and optionally a host URL and API version.

        Attributes:
        - api_key (str): User's API key.
        - host (str): Base URL for the API endpoints. Defaults to localhost.
        - version (str): API version. Defaults to "v1".
        - timeout (float): Request timeout in seconds. Defaults to 120.0 seconds.
        """
        self.api_key = api_key
        self.base_url = f"{host}/api/{version}/"
        self.headers = {
            'Content-Type': 'application/json'
        }
        self.timeout = timeout



# -*- coding: utf-8 -*-
"""
OpenAlgo Technical Indicators - Trend Indicators
"""

import numpy as np
import pandas as pd
from openalgo.numba_shim import jit
from typing import Union, Tuple, Optional
from .base import BaseIndicator
from .utils import sma, ema, highest, lowest, vwma_optimized, kama_optimized, atr_wilder


class SMA(BaseIndicator):
    """
    Simple Moving Average
    
    The SMA is calculated by adding the closing prices of a security for a period 
    and then dividing this total by the number of time periods.
    
    Formula: SMA = (P1 + P2 + ... + Pn) / n
    """
    
    def __init__(self):
        super().__init__("SMA")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_sma(data: np.ndarray, period: int) -> np.ndarray:
        """Numba optimized SMA calculation"""
        result = np.empty_like(data)
        result[:period-1] = np.nan
        
        # Calculate first SMA value
        sum_val = 0.0
        for i in range(period):
            sum_val += data[i]
        result[period-1] = sum_val / period
        
        # Calculate remaining values using rolling window
        for i in range(period, len(data)):
            sum_val = sum_val - data[i-period] + data[i]
            result[i] = sum_val / period
        
        return result
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Simple Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            SMA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        result = self._calculate_sma(validated_data, period)
        return self.format_output(result, input_type, index)


class EMA(BaseIndicator):
    """
    Exponential Moving Average
    
    The EMA gives more weight to recent prices, making it more responsive to new information.
    
    Formula: EMA = (Close - Previous EMA) × Multiplier + Previous EMA
    Where: Multiplier = 2 / (Period + 1)
    """
    
    def __init__(self):
        super().__init__("EMA")
    
    # Removed redundant EMA calculation - using consolidated utility
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Exponential Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            EMA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        result = ema(validated_data, period)
        return self.format_output(result, input_type, index)


class WMA(BaseIndicator):
    """
    Weighted Moving Average
    
    The WMA assigns greater weight to recent data points.
    
    Formula: WMA = (P1×1 + P2×2 + ... + Pn×n) / (1 + 2 + ... + n)
    """
    
    def __init__(self):
        super().__init__("WMA")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_wma(data: np.ndarray, period: int) -> np.ndarray:
        """Numba optimized WMA calculation"""
        result = np.empty_like(data)
        result[:period-1] = np.nan
        
        # Calculate weight sum: 1 + 2 + ... + period
        weight_sum = period * (period + 1) // 2
        
        for i in range(period-1, len(data)):
            weighted_sum = 0.0
            for j in range(period):
                weight = j + 1
                weighted_sum += data[i - period + 1 + j] * weight
            result[i] = weighted_sum / weight_sum
        
        return result
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Weighted Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            WMA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        result = self._calculate_wma(validated_data, period)
        return self.format_output(result, input_type, index)


class DEMA(BaseIndicator):
    """
    Double Exponential Moving Average
    
    DEMA attempts to reduce the lag associated with traditional moving averages.
    
    Formula: DEMA = 2 × EMA(n) - EMA(EMA(n))
    """
    
    def __init__(self):
        super().__init__("DEMA")
        self._ema = EMA()
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Double Exponential Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            DEMA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        
        # Calculate first EMA using consolidated utility
        ema1_data = ema(validated_data, period)
        
        # For second EMA, skip NaN values from first EMA
        # Find first valid index in ema1
        first_valid = period - 1
        if first_valid >= len(ema1_data):
            result = np.full_like(validated_data, np.nan)
            return self.format_output(result, input_type, index)
        
        # Extract valid portion for second EMA calculation
        valid_ema1 = ema1_data[first_valid:]
        ema2_partial = ema(valid_ema1, period)
        
        # Reconstruct full ema2 array
        ema2_data = np.full_like(validated_data, np.nan)
        second_valid_start = first_valid + period - 1
        if second_valid_start < len(ema2_data):
            valid_length = min(len(ema2_partial[period-1:]), len(ema2_data) - second_valid_start)
            ema2_data[second_valid_start:second_valid_start + valid_length] = ema2_partial[period-1:period-1 + valid_length]
        
        # DEMA = 2 * EMA - EMA(EMA)
        result = 2 * ema1_data - ema2_data
        return self.format_output(result, input_type, index)


class TEMA(BaseIndicator):
    """
    Triple Exponential Moving Average
    
    TEMA further reduces lag compared to DEMA.
    
    Formula: TEMA = 3×EMA(n) - 3×EMA(EMA(n)) + EMA(EMA(EMA(n)))
    """
    
    def __init__(self):
        super().__init__("TEMA")
        self._ema = EMA()
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Triple Exponential Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            TEMA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        
        # Calculate first EMA
        ema1_data = ema(validated_data, period)
        
        # Calculate second EMA from valid portion of first EMA
        first_valid = period - 1
        if first_valid >= len(ema1_data):
            result = np.full_like(validated_data, np.nan)
            return self.format_output(result, input_type, index)
        
        valid_ema1 = ema1_data[first_valid:]
        ema2_partial = ema(valid_ema1, period)
        
        # Reconstruct full ema2 array
        ema2_data = np.full_like(validated_data, np.nan)
        second_valid_start = first_valid + period - 1
        if second_valid_start < len(ema2_data):
            valid_length = min(len(ema2_partial[period-1:]), len(ema2_data) - second_valid_start)
            ema2_data[second_valid_start:second_valid_start + valid_length] = ema2_partial[period-1:period-1 + valid_length]
        
        # Calculate third EMA from valid portion of second EMA
        ema3_data = np.full_like(validated_data, np.nan)
        if second_valid_start < len(ema2_data):
            valid_ema2 = ema2_data[second_valid_start:]
            # Remove NaN values
            valid_ema2_clean = valid_ema2[~np.isnan(valid_ema2)]
            if len(valid_ema2_clean) >= period:
                ema3_partial = ema(valid_ema2_clean, period)
                third_valid_start = second_valid_start + period - 1
                if third_valid_start < len(ema3_data):
                    valid_length = min(len(ema3_partial[period-1:]), len(ema3_data) - third_valid_start)
                    ema3_data[third_valid_start:third_valid_start + valid_length] = ema3_partial[period-1:period-1 + valid_length]
        
        # TEMA = 3 * EMA - 3 * EMA(EMA) + EMA(EMA(EMA))
        result = 3 * ema1_data - 3 * ema2_data + ema3_data
        return self.format_output(result, input_type, index)


@jit(nopython=True)
def _calculate_atr(high: np.ndarray, low: np.ndarray, close: np.ndarray, period: int) -> np.ndarray:
    """Calculate Average True Range"""
    n = len(high)
    tr = np.empty(n)
    atr = np.empty(n)
    
    # First TR value
    tr[0] = high[0] - low[0]
    
    # Calculate True Range
    for i in range(1, n):
        hl = high[i] - low[i]
        hc = abs(high[i] - close[i-1])
        lc = abs(low[i] - close[i-1])
        tr[i] = max(hl, hc, lc)
    
    # Calculate ATR
    atr[:period-1] = np.nan
    
    # Initial ATR is simple average
    sum_tr = 0.0
    for i in range(period):
        sum_tr += tr[i]
    atr[period-1] = sum_tr / period
    
    # Subsequent ATR values use smoothed average
    for i in range(period, n):
        atr[i] = (atr[i-1] * (period - 1) + tr[i]) / period
    
    return atr


class Supertrend(BaseIndicator):
    """
    Supertrend Indicator
    
    The Supertrend indicator is a trend-following indicator that uses ATR (Average True Range) 
    to calculate dynamic support and resistance levels.
    
    Formula:
    Basic Upper Band = (HIGH + LOW) / 2 + Multiplier × ATR
    Basic Lower Band = (HIGH + LOW) / 2 - Multiplier × ATR
    """
    
    def __init__(self):
        super().__init__("Supertrend")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_supertrend(high: np.ndarray, low: np.ndarray, close: np.ndarray, 
                             period: int, multiplier: float) -> Tuple[np.ndarray, np.ndarray]:
        """Numba optimized Supertrend calculation - matches TradingView Pine Script logic"""
        n = len(close)
        
        # Calculate ATR
        atr = _calculate_atr(high, low, close, period)
        
        # Calculate basic bands (src = hl2 in Pine Script)
        hl_avg = (high + low) / 2.0
        upper_band = hl_avg + multiplier * atr
        lower_band = hl_avg - multiplier * atr
        
        # Initialize arrays
        final_upper = np.full(n, np.nan)
        final_lower = np.full(n, np.nan)
        supertrend = np.full(n, np.nan)
        direction = np.full(n, np.nan)
        
        first_valid = period - 1
        if first_valid < 0 or first_valid >= n:
            return supertrend, direction
        
        # Initialize first valid values
        final_upper[first_valid] = upper_band[first_valid]
        final_lower[first_valid] = lower_band[first_valid]
        # Pine Script: if na(atr[1]) _direction := 1 (first bar is downtrend)
        direction[first_valid] = 1.0  # downtrend (red in Pine Script)
        supertrend[first_valid] = final_upper[first_valid]
        
        # Pine Script logic for subsequent bars
        for i in range(first_valid + 1, n):
            # Final lower band: lowerBand > prevLowerBand or close[1] < prevLowerBand ? lowerBand : prevLowerBand
            if lower_band[i] > final_lower[i-1] or close[i-1] < final_lower[i-1]:
                final_lower[i] = lower_band[i]
            else:
                final_lower[i] = final_lower[i-1]
            
            # Final upper band: upperBand < prevUpperBand or close[1] > prevUpperBand ? upperBand : prevUpperBand  
            if upper_band[i] < final_upper[i-1] or close[i-1] > final_upper[i-1]:
                final_upper[i] = upper_band[i]
            else:
                final_upper[i] = final_upper[i-1]
            
            # Direction logic (Pine Script)
            # if prevSuperTrend == prevUpperBand
            #     _direction := close > upperBand ? -1 : 1
            # else
            #     _direction := close < lowerBand ? 1 : -1
            if supertrend[i-1] == final_upper[i-1]:
                # Previous was upper band (downtrend)
                if close[i] > final_upper[i]:
                    direction[i] = -1.0  # Change to uptrend (green)
                else:
                    direction[i] = 1.0   # Continue downtrend (red)
            else:
                # Previous was lower band (uptrend)
                if close[i] < final_lower[i]:
                    direction[i] = 1.0   # Change to downtrend (red)
                else:
                    direction[i] = -1.0  # Continue uptrend (green)
            
            # Supertrend assignment: _direction == -1 ? lowerBand : upperBand
            if direction[i] == -1.0:  # uptrend (green)
                supertrend[i] = final_lower[i]
            else:  # downtrend (red)
                supertrend[i] = final_upper[i]
                
        return supertrend, direction
    
    def calculate(self, high: Union[np.ndarray, pd.Series, list],
                 low: Union[np.ndarray, pd.Series, list],
                 close: Union[np.ndarray, pd.Series, list],
                 period: int = 10, multiplier: float = 3.0) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]:
        """
        Calculate Supertrend Indicator
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        period : int, default=10
            ATR period
        multiplier : float, default=3.0
            ATR multiplier
            
        Returns:
        --------
        Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]
            (supertrend values, direction values) in the same format as input
            Direction: -1 for uptrend (green), 1 for downtrend (red) - matches TradingView
        """
        high_data, input_type, index = self.validate_input(high)
        low_data, _, _ = self.validate_input(low)
        close_data, _, _ = self.validate_input(close)
        
        # Align arrays
        high_data, low_data, close_data = self.align_arrays(high_data, low_data, close_data)
        self.validate_period(period, len(close_data))
        
        if multiplier <= 0:
            raise ValueError(f"Multiplier must be positive, got {multiplier}")
        
        supertrend_result, direction_result = self._calculate_supertrend(high_data, low_data, close_data, period, multiplier)
        return self.format_multiple_outputs((supertrend_result, direction_result), input_type, index)


class Ichimoku(BaseIndicator):
    """
    Ichimoku Cloud - matches TradingView exactly
    
    The Ichimoku Cloud is a comprehensive indicator that defines support and resistance, 
    identifies trend direction, gauges momentum, and provides trading signals.
    
    TradingView Components:
    - Conversion Line: donchian(conversionPeriods) = avg(highest(9), lowest(9))
    - Base Line: donchian(basePeriods) = avg(highest(26), lowest(26))  
    - Leading Span A: avg(conversionLine, baseLine), offset +displacement-1
    - Leading Span B: donchian(laggingSpan2Periods) = avg(highest(52), lowest(52)), offset +displacement-1
    - Lagging Span: close, offset -displacement+1
    """
    
    def __init__(self):
        super().__init__("Ichimoku")
    
    @staticmethod
    def _calculate_ichimoku_tv(high: np.ndarray, low: np.ndarray, close: np.ndarray,
                              conversion_periods: int = 9, base_periods: int = 26,
                              lagging_span2_periods: int = 52, displacement: int = 26) -> Tuple[np.ndarray, ...]:
        """
        Calculate Ichimoku using exact TradingView logic
        TradingView formula:
        donchian(len) => math.avg(ta.lowest(len), ta.highest(len))
        conversionLine = donchian(conversionPeriods)
        baseLine = donchian(basePeriods)
        leadLine1 = math.avg(conversionLine, baseLine)
        leadLine2 = donchian(laggingSpan2Periods)
        """
        n = len(close)
        
        # Donchian function: avg(highest, lowest)
        def donchian(data_high, data_low, period):
            high_values = highest(data_high, period)
            low_values = lowest(data_low, period)
            return (high_values + low_values) / 2.0
        
        # Calculate components using TradingView names and logic
        conversion_line = donchian(high, low, conversion_periods)  # Tenkan-sen
        base_line = donchian(high, low, base_periods)             # Kijun-sen
        lead_line1 = (conversion_line + base_line) / 2.0          # Senkou Span A
        lead_line2 = donchian(high, low, lagging_span2_periods)   # Senkou Span B
        
        # Apply TradingView offset logic
        # TradingView: offset = displacement - 1 for leading spans
        # TradingView: offset = -displacement + 1 for lagging span
        
        # Leading Span A: offset = displacement - 1 (shift forward)
        leading_span_a = np.full(n, np.nan)
        offset_a = displacement - 1
        if offset_a > 0 and offset_a < n:
            leading_span_a[offset_a:] = lead_line1[:-offset_a]
        elif offset_a == 0:
            leading_span_a = lead_line1.copy()
        
        # Leading Span B: offset = displacement - 1 (shift forward)  
        leading_span_b = np.full(n, np.nan)
        offset_b = displacement - 1
        if offset_b > 0 and offset_b < n:
            leading_span_b[offset_b:] = lead_line2[:-offset_b]
        elif offset_b == 0:
            leading_span_b = lead_line2.copy()
        
        # Lagging Span: offset = -displacement + 1 (shift backward)
        lagging_span = np.full(n, np.nan)
        offset_lag = -displacement + 1
        if offset_lag < 0:
            lag_shift = abs(offset_lag)
            if lag_shift < n:
                lagging_span[:-lag_shift] = close[lag_shift:]
        elif offset_lag > 0:
            if offset_lag < n:
                lagging_span[offset_lag:] = close[:-offset_lag]
        else:  # offset_lag == 0
            lagging_span = close.copy()
        
        return conversion_line, base_line, leading_span_a, leading_span_b, lagging_span
    
    def calculate(self, high: Union[np.ndarray, pd.Series, list],
                 low: Union[np.ndarray, pd.Series, list],
                 close: Union[np.ndarray, pd.Series, list],
                 conversion_periods: int = 9, base_periods: int = 26,
                 lagging_span2_periods: int = 52, displacement: int = 26) -> Union[Tuple[np.ndarray, ...], Tuple[pd.Series, ...]]:
        """
        Calculate Ichimoku Cloud - matches TradingView exactly
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        conversion_periods : int, default=9
            Conversion Line Length (TradingView: conversionPeriods)
        base_periods : int, default=26
            Base Line Length (TradingView: basePeriods)
        lagging_span2_periods : int, default=52
            Leading Span B Length (TradingView: laggingSpan2Periods)
        displacement : int, default=26
            Lagging Span displacement (TradingView: displacement)
            
        Returns:
        --------
        Union[Tuple[np.ndarray, ...], Tuple[pd.Series, ...]]
            (conversion_line, base_line, leading_span_a, leading_span_b, lagging_span) in the same format as input
        """
        high_data, input_type, index = self.validate_input(high)
        low_data, _, _ = self.validate_input(low)
        close_data, _, _ = self.validate_input(close)
        
        # Align arrays
        high_data, low_data, close_data = self.align_arrays(high_data, low_data, close_data)
        
        # Validate periods (use TradingView parameter names)
        for period, name in [(conversion_periods, "conversion_periods"), 
                            (base_periods, "base_periods"), 
                            (lagging_span2_periods, "lagging_span2_periods")]:
            if period <= 0:
                raise ValueError(f"{name} must be positive, got {period}")
        
        results = self._calculate_ichimoku_tv(high_data, low_data, close_data, conversion_periods, 
                                             base_periods, lagging_span2_periods, displacement)
        return self.format_multiple_outputs(results, input_type, index)


class HMA(BaseIndicator):
    """
    Hull Moving Average
    
    The Hull Moving Average (HMA) attempts to minimize lag and improve smoothing.
    
    Formula: HMA = WMA(2 × WMA(n/2) - WMA(n), sqrt(n))
    """
    
    def __init__(self):
        super().__init__("HMA")
        self._wma = WMA()
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Hull Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            HMA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        
        # Step 1: Calculate WMA(n/2)
        wma_half = self._wma.calculate(validated_data, period // 2)
        
        # Step 2: Calculate WMA(n)
        wma_full = self._wma.calculate(validated_data, period)
        
        # Step 3: Calculate 2 * WMA(n/2) - WMA(n)
        diff = 2 * wma_half - wma_full
        
        # Step 4: Calculate HMA = WMA(diff, sqrt(n))
        sqrt_period = int(np.sqrt(period))
        result = self._wma.calculate(diff, sqrt_period)
        
        return self.format_output(result, input_type, index)


class VWMA(BaseIndicator):
    """
    Volume Weighted Moving Average
    
    VWMA gives more weight to periods with higher volume.
    
    Formula: VWMA = Σ(Price × Volume) / Σ(Volume)
    """
    
    def __init__(self):
        super().__init__("VWMA")
    
    @staticmethod
    def _calculate_vwma(data: np.ndarray, volume: np.ndarray, period: int) -> np.ndarray:
        """O(n) optimized VWMA calculation using utils"""
        return vwma_optimized(data, volume, period)
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list],
                 volume: Union[np.ndarray, pd.Series, list],
                 period: int) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Volume Weighted Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        volume : Union[np.ndarray, pd.Series, list]
            Volume data
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            VWMA values in the same format as input
        """
        data_validated, input_type, index = self.validate_input(data)
        volume_validated, _, _ = self.validate_input(volume)
        data_validated, volume_validated = self.align_arrays(data_validated, volume_validated)
        self.validate_period(period, len(data_validated))
        
        result = self._calculate_vwma(data_validated, volume_validated, period)
        return self.format_output(result, input_type, index)


class ALMA(BaseIndicator):
    """
    Arnaud Legoux Moving Average
    
    ALMA is a technical analysis indicator that combines the features of SMA and EMA.
    
    Formula: ALMA = Σ(w[i] × data[i]) / Σ(w[i])
    Where: w[i] = exp(-((i - m)^2) / (2 × s^2))
    """
    
    def __init__(self):
        super().__init__("ALMA")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_alma(data: np.ndarray, period: int, offset: float, sigma: float) -> np.ndarray:
        """Numba optimized ALMA calculation"""
        n = len(data)
        result = np.full(n, np.nan)
        
        # Calculate weights
        m = offset * (period - 1)
        s = period / sigma
        
        weights = np.empty(period)
        for i in range(period):
            weights[i] = np.exp(-((i - m) ** 2) / (2 * s * s))
        
        # Normalize weights
        weight_sum = np.sum(weights)
        weights = weights / weight_sum
        
        # Calculate ALMA
        for i in range(period - 1, n):
            alma_sum = 0.0
            for j in range(period):
                alma_sum += weights[j] * data[i - period + 1 + j]
            result[i] = alma_sum
        
        return result
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list],
                 period: int = 21, offset: float = 0.85, sigma: float = 6.0) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Arnaud Legoux Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int, default=21
            Number of periods for the moving average
        offset : float, default=0.85
            Phase offset (0 to 1)
        sigma : float, default=6.0
            Smoothing factor
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            ALMA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        
        if not 0 <= offset <= 1:
            raise ValueError(f"Offset must be between 0 and 1, got {offset}")
        if sigma <= 0:
            raise ValueError(f"Sigma must be positive, got {sigma}")
        
        result = self._calculate_alma(validated_data, period, offset, sigma)
        return self.format_output(result, input_type, index)


class KAMA(BaseIndicator):
    """
    Kaufman's Adaptive Moving Average - matches TradingView exactly
    
    KAMA is a moving average designed to account for market noise or volatility.
    
    TradingView Formula:
    mom = abs(change(src, length))
    volatility = sum(abs(change(src)), length)  
    er = volatility != 0 ? mom / volatility : 0
    fastAlpha = 2 / (fastLength + 1)
    slowAlpha = 2 / (slowLength + 1)
    alpha = pow(er * (fastAlpha - slowAlpha) + slowAlpha, 2)
    kama = alpha * src + (1 - alpha) * nz(kama[1], src)
    """
    
    def __init__(self):
        super().__init__("KAMA")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_kama_tv(data: np.ndarray, length: int, fast_length: int, slow_length: int) -> np.ndarray:
        """
        Calculate KAMA using exact TradingView logic
        """
        n = len(data)
        result = np.full(n, np.nan)
        
        if n < length + 1:
            return result
        
        # TradingView alpha calculations
        fast_alpha = 2.0 / (fast_length + 1)
        slow_alpha = 2.0 / (slow_length + 1)
        
        # Calculate for each period
        for i in range(length, n):
            # mom = abs(change(src, length)) -> abs(data[i] - data[i-length])
            mom = abs(data[i] - data[i - length])
            
            # volatility = sum(abs(change(src)), length) -> sum of absolute changes over length
            volatility = 0.0
            for j in range(i - length + 1, i + 1):
                if j > 0:
                    volatility += abs(data[j] - data[j - 1])
            
            # Efficiency Ratio: er = volatility != 0 ? mom / volatility : 0
            if volatility != 0:
                er = mom / volatility
            else:
                er = 0.0
            
            # alpha = pow(er * (fastAlpha - slowAlpha) + slowAlpha, 2)
            alpha = (er * (fast_alpha - slow_alpha) + slow_alpha) ** 2
            
            # kama = alpha * src + (1 - alpha) * nz(kama[1], src)
            # For first calculation, use current price as previous KAMA (nz logic)
            if i == length or np.isnan(result[i - 1]):
                prev_kama = data[i]
            else:
                prev_kama = result[i - 1]
            
            result[i] = alpha * data[i] + (1 - alpha) * prev_kama
        
        return result
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list],
                 length: int = 14, fast_length: int = 2, slow_length: int = 30) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Kaufman's Adaptive Moving Average - matches TradingView exactly
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Source data (typically closing prices)
        length : int, default=14
            Length for efficiency ratio calculation (TradingView default)
        fast_length : int, default=2
            Fast EMA Length (TradingView: fastLength)
        slow_length : int, default=30
            Slow EMA Length (TradingView: slowLength)
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            KAMA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(length + 1, len(validated_data))
        
        if length <= 0:
            raise ValueError(f"Length must be positive, got {length}")
        if fast_length <= 0:
            raise ValueError(f"Fast length must be positive, got {fast_length}")
        if slow_length <= 0:
            raise ValueError(f"Slow length must be positive, got {slow_length}")
        if fast_length >= slow_length:
            raise ValueError("Fast length must be less than slow length")
        
        result = self._calculate_kama_tv(validated_data, length, fast_length, slow_length)
        return self.format_output(result, input_type, index)


class ZLEMA(BaseIndicator):
    """
    Zero Lag Exponential Moving Average
    
    ZLEMA is an EMA that attempts to minimize lag by using price momentum.
    
    Formula: ZLEMA = EMA(2 × Price - Price[lag])
    Where: lag = (period - 1) / 2
    """
    
    def __init__(self):
        super().__init__("ZLEMA")
        self._ema = EMA()
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Zero Lag Exponential Moving Average using O(n) algorithm
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            ZLEMA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        
        result = self._calculate_zlema_optimized(validated_data, period)
        return self.format_output(result, input_type, index)
    
    @staticmethod
    @jit(nopython=True, cache=True)
    def _calculate_zlema_optimized(data: np.ndarray, period: int) -> np.ndarray:
        """O(n) optimized ZLEMA calculation using consolidated EMA utility"""
        n = len(data)
        lag = (period - 1) // 2
        
        # Create adjusted data using vectorized operations where possible
        adjusted_data = np.empty(n)
        adjusted_data[:lag] = data[:lag]
        
        # Vectorized calculation for the main portion
        for i in range(lag, n):
            adjusted_data[i] = 2 * data[i] - data[i - lag]
        
        # Use the optimized EMA utility
        return ema(adjusted_data, period)


class T3(BaseIndicator):
    """
    T3 Moving Average
    
    T3 is a type of moving average which is the result of applying EMA three times.
    
    Formula: T3 = GD(GD(GD(data)))
    Where: GD = Generalized DEMA
    """
    
    def __init__(self):
        super().__init__("T3")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_gd(data: np.ndarray, period: int, v_factor: float) -> np.ndarray:
        """Calculate Generalized DEMA"""
        alpha = 2.0 / (period + 1)
        
        # First EMA
        ema1 = np.empty_like(data)
        ema1[0] = data[0]
        for i in range(1, len(data)):
            ema1[i] = alpha * data[i] + (1 - alpha) * ema1[i - 1]
        
        # Second EMA
        ema2 = np.empty_like(data)
        ema2[0] = ema1[0]
        for i in range(1, len(data)):
            ema2[i] = alpha * ema1[i] + (1 - alpha) * ema2[i - 1]
        
        # Generalized DEMA
        gd = (1 + v_factor) * ema1 - v_factor * ema2
        return gd
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list],
                 period: int = 21, v_factor: float = 0.7) -> Union[np.ndarray, pd.Series]:
        """
        Calculate T3 Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int, default=21
            Number of periods for the moving average
        v_factor : float, default=0.7
            Volume factor for T3 calculation
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            T3 values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        
        # Apply GD three times
        gd1 = self._calculate_gd(validated_data, period, v_factor)
        gd2 = self._calculate_gd(gd1, period, v_factor)
        result = self._calculate_gd(gd2, period, v_factor)
        
        return self.format_output(result, input_type, index)


class FRAMA(BaseIndicator):
    """
    Fractal Adaptive Moving Average
    
    FRAMA adjusts its smoothing based on the fractal dimension of the price series.
    
    Formula: FRAMA = (2 - D) × Price + (D - 1) × FRAMA[prev]
    Where: D = fractal dimension
    """
    
    def __init__(self):
        super().__init__("FRAMA")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_frama_tv(high: np.ndarray, low: np.ndarray, period: int) -> np.ndarray:
        """
        Calculate FRAMA using exact TradingView logic
        
        This matches the TradingView FRAMA Channel implementation exactly:
        1. Uses hl2 as price source
        2. Calculates N1, N2, N3 exactly as in TV script
        3. Uses TV fractal dimension and alpha formula
        4. Applies SMA smoothing at the end
        """
        n = len(high)
        filt = np.full(n, np.nan)
        
        # Calculate hl2 (price source)
        price = (high + low) / 2.0
        
        # Set initial value
        if n > 0:
            filt[0] = price[0]
        
        for i in range(1, n):
            if i >= period:
                # Calculate N3 (range over full period N)
                N3 = (np.max(high[i - period + 1:i + 1]) - np.min(low[i - period + 1:i + 1])) / period
                
                # Calculate N1 (first half)
                half_period = period // 2
                HH = high[i]
                LL = low[i]
                
                # Loop to find HH and LL for first half
                for count in range(0, half_period):
                    idx = i - count
                    if idx >= 0:
                        if high[idx] > HH:
                            HH = high[idx]
                        if low[idx] < LL:
                            LL = low[idx]
                
                N1 = (HH - LL) / half_period
                
                # Calculate N2 (second half)
                HH = high[i - half_period] if i - half_period >= 0 else high[0]
                LL = low[i - half_period] if i - half_period >= 0 else low[0]
                
                # Loop to find HH and LL for second half
                for count in range(half_period, period):
                    idx = i - count
                    if idx >= 0:
                        if high[idx] > HH:
                            HH = high[idx]
                        if low[idx] < LL:
                            LL = low[idx]
                
                N2 = (HH - LL) / half_period
                
                # Calculate fractal dimension (TradingView formula)
                if N1 > 0 and N2 > 0 and N3 > 0:
                    Dimen = (np.log(N1 + N2) - np.log(N3)) / np.log(2)
                else:
                    Dimen = 1.0
                
                # Calculate alpha (TradingView formula)
                alpha = np.exp(-4.6 * (Dimen - 1))
                alpha = max(min(alpha, 1.0), 0.01)  # Clamp between 0.01 and 1
                
                # Calculate FRAMA filtered value
                filt[i] = alpha * price[i] + (1 - alpha) * filt[i - 1]
            else:
                # For initial periods before we have enough data, use price
                filt[i] = price[i]
        
        # Apply SMA(5) smoothing as in TradingView script
        smoothed = np.full(n, np.nan)
        
        for i in range(n):
            if i < period + 1:
                # TradingView: ta.sma((bar_index < N + 1) ? price : Filt, 5)
                # For early periods, use price directly
                smoothed[i] = price[i]
            else:
                # Use the filtered value
                if i >= 4:  # Need 5 periods for SMA
                    window_start = max(0, i - 4)
                    sma_sum = 0.0
                    valid_count = 0
                    for j in range(window_start, i + 1):
                        if not np.isnan(filt[j]):
                            sma_sum += filt[j]
                            valid_count += 1
                    if valid_count > 0:
                        smoothed[i] = sma_sum / valid_count
                    else:
                        smoothed[i] = filt[i]
                else:
                    smoothed[i] = filt[i]
        
        return smoothed
    
    def calculate(self, high: Union[np.ndarray, pd.Series, list],
                 low: Union[np.ndarray, pd.Series, list],
                 period: int = 26) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Fractal Adaptive Moving Average - matches TradingView exactly
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices  
        period : int, default=26
            Number of periods for FRAMA calculation (TradingView default)
            Must be even number, minimum 2
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            FRAMA values in the same format as input
        """
        high_data, input_type, index = self.validate_input(high)
        low_data, _, _ = self.validate_input(low)
        
        high_data, low_data = self.align_arrays(high_data, low_data)
        self.validate_period(period, len(high_data))
        
        if period < 2:
            raise ValueError("Period must be at least 2 for FRAMA calculation")
        
        if period % 2 != 0:
            raise ValueError("Period must be even for FRAMA calculation")
        
        result = self._calculate_frama_tv(high_data, low_data, period)
        return self.format_output(result, input_type, index)


class ChandeKrollStop(BaseIndicator):
    """
    Chande Kroll Stop
    
    The Chande Kroll Stop is a trailing stop-loss technique that combines 
    ATR-based calculations with highest highs and lowest lows.
    
    Formula:
    Long Stop = Highest(High, period) - ATR_multiplier * ATR(period)
    Short Stop = Lowest(Low, period) + ATR_multiplier * ATR(period)
    """
    
    def __init__(self):
        super().__init__("Chande Kroll Stop")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_atr(high: np.ndarray, low: np.ndarray, close: np.ndarray, period: int) -> np.ndarray:
        """Calculate ATR"""
        n = len(close)
        tr = np.empty(n)
        
        # Calculate True Range
        tr[0] = high[0] - low[0]
        for i in range(1, n):
            tr[i] = max(high[i] - low[i],
                       abs(high[i] - close[i-1]),
                       abs(low[i] - close[i-1]))
        
        # Calculate ATR using exponential moving average
        atr = np.empty(n)
        atr[0] = tr[0]
        alpha = 1.0 / period
        
        for i in range(1, n):
            atr[i] = alpha * tr[i] + (1 - alpha) * atr[i-1]
        
        return atr
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_chande_kroll_stop(high: np.ndarray, low: np.ndarray, close: np.ndarray,
                                   p: int, x: float, q: int) -> Tuple[np.ndarray, np.ndarray]:
        """Numba optimized Chande Kroll Stop calculation - matches TradingView Pine Script"""
        n = len(close)
        long_stop = np.full(n, np.nan)
        short_stop = np.full(n, np.nan)
        
        # Calculate ATR using Wilder's method (matches TradingView ta.atr)
        atr = atr_wilder(high, low, close, p)
        
        # Step 1: Calculate first_high_stop and first_low_stop
        first_high_stop = np.full(n, np.nan)
        first_low_stop = np.full(n, np.nan)
        
        for i in range(p - 1, n):
            # Calculate highest high and lowest low over period p
            highest_high = np.max(high[i - p + 1:i + 1])
            lowest_low = np.min(low[i - p + 1:i + 1])
            
            # TradingView: first_high_stop = ta.highest(high, p) - x * ta.atr(p)
            first_high_stop[i] = highest_high - x * atr[i]
            # TradingView: first_low_stop = ta.lowest(low, p) + x * ta.atr(p)
            first_low_stop[i] = lowest_low + x * atr[i]
        
        # Step 2: Calculate final stops using highest/lowest of first stops over period q
        # TradingView: stop_short = ta.highest(first_high_stop, q)
        # TradingView: stop_long = ta.lowest(first_low_stop, q)
        start_idx = p + q - 2  # Need both p and q periods to start
        
        for i in range(start_idx, n):
            # Calculate highest of first_high_stop over period q
            if i >= q - 1:
                q_start = max(0, i - q + 1)
                # Find highest non-NaN value in the q-period window
                window_high = first_high_stop[q_start:i + 1]
                valid_high = window_high[~np.isnan(window_high)]
                if len(valid_high) > 0:
                    short_stop[i] = np.max(valid_high)
                
                # Find lowest non-NaN value in the q-period window  
                window_low = first_low_stop[q_start:i + 1]
                valid_low = window_low[~np.isnan(window_low)]
                if len(valid_low) > 0:
                    long_stop[i] = np.min(valid_low)
        
        return long_stop, short_stop
    
    def calculate(self, high: Union[np.ndarray, pd.Series, list],
                 low: Union[np.ndarray, pd.Series, list],
                 close: Union[np.ndarray, pd.Series, list],
                 p: int = 10, x: float = 1.0, q: int = 9) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]:
        """
        Calculate Chande Kroll Stop - matches TradingView exactly
        
        Parameters:
        -----------
        high : Union[np.ndarray, pd.Series, list]
            High prices
        low : Union[np.ndarray, pd.Series, list]
            Low prices
        close : Union[np.ndarray, pd.Series, list]
            Closing prices
        p : int, default=10
            ATR Length (period for ATR and highest/lowest calculation)
        x : float, default=1.0
            ATR Coefficient (multiplier for ATR)
        q : int, default=9
            Stop Length (period for smoothing the stops)
            
        Returns:
        --------
        Union[Tuple[np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series]]
            (stop_long, stop_short) in the same format as input
            - stop_long: lowest of first_low_stop over q periods
            - stop_short: highest of first_high_stop over q periods
        """
        high_data, input_type, index = self.validate_input(high)
        low_data, _, _ = self.validate_input(low)
        close_data, _, _ = self.validate_input(close)
        
        high_data, low_data, close_data = self.align_arrays(high_data, low_data, close_data)
        self.validate_period(p, len(close_data))
        self.validate_period(q, len(close_data))
        
        if x <= 0:
            raise ValueError(f"ATR coefficient (x) must be positive, got {x}")
        
        long_stop, short_stop = self._calculate_chande_kroll_stop(high_data, low_data, close_data, p, x, q)
        
        results = (long_stop, short_stop)
        return self.format_multiple_outputs(results, input_type, index)


class TRIMA(BaseIndicator):
    """
    Triangular Moving Average
    
    TRIMA is a double-smoothed moving average that applies SMA twice.
    
    Formula: TRIMA = SMA(SMA(Close, n), n)
    Where n = (period + 1) / 2
    """
    
    def __init__(self):
        super().__init__("TRIMA")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_trima(data: np.ndarray, period: int) -> np.ndarray:
        """Numba optimized TRIMA calculation"""
        n = len(data)
        result = np.full(n, np.nan)
        
        # First, calculate n for SMA
        n1 = (period + 1) // 2
        n2 = period - n1 + 1
        
        # First SMA
        first_sma = np.full(n, np.nan)
        for i in range(n1 - 1, n):
            first_sma[i] = np.mean(data[i - n1 + 1:i + 1])
        
        # Second SMA on first SMA
        for i in range(n1 + n2 - 2, n):
            if not np.isnan(first_sma[i]):
                window_start = max(0, i - n2 + 1)
                window = first_sma[window_start:i + 1]
                valid_values = window[~np.isnan(window)]
                if len(valid_values) >= n2:
                    result[i] = np.mean(valid_values[-n2:])
        
        return result
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Triangular Moving Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the moving average
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            TRIMA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        result = self._calculate_trima(validated_data, period)
        return self.format_output(result, input_type, index)


class McGinley(BaseIndicator):
    """
    McGinley Dynamic
    
    A moving average that adjusts automatically for market speed changes.
    
    Formula: MD[i] = MD[i-1] + (Close[i] - MD[i-1]) / (N * (Close[i]/MD[i-1])^4)
    """
    
    def __init__(self):
        super().__init__("McGinley Dynamic")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_mcginley(data: np.ndarray, period: int) -> np.ndarray:
        """Numba optimized McGinley Dynamic calculation"""
        n = len(data)
        result = np.full(n, np.nan)
        
        if n < period:
            return result
        
        # Initialize with SMA
        result[period - 1] = np.mean(data[:period])
        
        for i in range(period, n):
            if result[i - 1] != 0:
                ratio = data[i] / result[i - 1]
                factor = period * (ratio ** 4)
                result[i] = result[i - 1] + (data[i] - result[i - 1]) / factor
            else:
                result[i] = data[i]
        
        return result
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list], period: int) -> Union[np.ndarray, pd.Series]:
        """
        Calculate McGinley Dynamic
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int
            Number of periods for the calculation
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            McGinley Dynamic values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        result = self._calculate_mcginley(validated_data, period)
        return self.format_output(result, input_type, index)


class VIDYA(BaseIndicator):
    """
    Variable Index Dynamic Average (VIDYA)
    
    VIDYA uses the Chande Momentum Oscillator (CMO) to adjust the smoothing constant of an EMA.
    
    Formula: VIDYA[i] = VIDYA[i-1] + alpha * |CMO[i]| / 100 * (Close[i] - VIDYA[i-1])
    """
    
    def __init__(self):
        super().__init__("VIDYA")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_cmo(data: np.ndarray, period: int) -> np.ndarray:
        """Calculate Chande Momentum Oscillator"""
        n = len(data)
        result = np.full(n, np.nan)
        
        # Calculate price changes
        changes = np.diff(data)
        
        for i in range(period, n):
            sum_up = 0.0
            sum_down = 0.0
            
            for j in range(period):
                change = changes[i - period + j]
                if change > 0:
                    sum_up += change
                elif change < 0:
                    sum_down += abs(change)
            
            total_movement = sum_up + sum_down
            if total_movement > 0:
                result[i] = 100 * (sum_up - sum_down) / total_movement
            else:
                result[i] = 0.0
        
        return result
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_vidya(data: np.ndarray, period: int, alpha: float) -> np.ndarray:
        """Numba optimized VIDYA calculation"""
        n = len(data)
        result = np.full(n, np.nan)
        
        # Calculate CMO inline
        cmo = np.full(n, np.nan)
        
        for i in range(period, n):
            gains = 0.0
            losses = 0.0
            
            for j in range(i - period + 1, i + 1):
                if j > 0:
                    diff = data[j] - data[j - 1]
                    if diff > 0:
                        gains += diff
                    elif diff < 0:
                        losses += abs(diff)
            
            if gains + losses != 0:
                cmo[i] = 100 * (gains - losses) / (gains + losses)
            else:
                cmo[i] = 0.0
        
        # Initialize with first valid data point
        result[period] = data[period]
        
        for i in range(period + 1, n):
            if not np.isnan(cmo[i]):
                smoothing_constant = alpha * abs(cmo[i]) / 100
                result[i] = result[i - 1] + smoothing_constant * (data[i] - result[i - 1])
            else:
                result[i] = result[i - 1]
        
        return result
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list], 
                 period: int = 14, alpha: float = 0.2) -> Union[np.ndarray, pd.Series]:
        """
        Calculate Variable Index Dynamic Average
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int, default=14
            Period for CMO calculation
        alpha : float, default=0.2
            Alpha factor for smoothing
            
        Returns:
        --------
        Union[np.ndarray, pd.Series]
            VIDYA values in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period + 1, len(validated_data))
        result = self._calculate_vidya(validated_data, period, alpha)
        return self.format_output(result, input_type, index)


class Alligator(BaseIndicator):
    """
    Alligator (Bill Williams)
    
    The Alligator consists of three smoothed moving averages:
    - Jaw (blue line): 13-period SMMA, shifted 8 bars forward
    - Teeth (red line): 8-period SMMA, shifted 5 bars forward  
    - Lips (green line): 5-period SMMA, shifted 3 bars forward
    """
    
    def __init__(self):
        super().__init__("Alligator")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_smma(data: np.ndarray, period: int) -> np.ndarray:
        """Calculate Smoothed Moving Average (SMMA)"""
        n = len(data)
        result = np.full(n, np.nan)
        
        if n < period:
            return result
        
        # Initialize with SMA
        result[period - 1] = np.mean(data[:period])
        
        # Calculate SMMA
        for i in range(period, n):
            result[i] = (result[i - 1] * (period - 1) + data[i]) / period
        
        return result
    
    @staticmethod
    @jit(nopython=True)
    def _shift_series(data: np.ndarray, shift: int) -> np.ndarray:
        """Shift series forward by specified periods"""
        n = len(data)
        result = np.full(n, np.nan)
        
        for i in range(shift, n):
            result[i] = data[i - shift]
        
        return result
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list],
                 jaw_period: int = 13, jaw_shift: int = 8,
                 teeth_period: int = 8, teeth_shift: int = 5,
                 lips_period: int = 5, lips_shift: int = 3) -> Union[Tuple[np.ndarray, np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series, pd.Series]]:
        """
        Calculate Alligator
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically (high + low) / 2)
        jaw_period : int, default=13
            Period for Jaw line
        jaw_shift : int, default=8
            Forward shift for Jaw line
        teeth_period : int, default=8
            Period for Teeth line
        teeth_shift : int, default=5
            Forward shift for Teeth line
        lips_period : int, default=5
            Period for Lips line
        lips_shift : int, default=3
            Forward shift for Lips line
            
        Returns:
        --------
        Union[Tuple[np.ndarray, np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series, pd.Series]]
            (jaw, teeth, lips) in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        
        # Calculate SMMA for each line
        jaw_smma = self._calculate_smma(validated_data, jaw_period)
        teeth_smma = self._calculate_smma(validated_data, teeth_period)
        lips_smma = self._calculate_smma(validated_data, lips_period)
        
        # Apply shifts
        jaw = self._shift_series(jaw_smma, jaw_shift)
        teeth = self._shift_series(teeth_smma, teeth_shift)
        lips = self._shift_series(lips_smma, lips_shift)
        
        results = (jaw, teeth, lips)
        return self.format_multiple_outputs(results, input_type, index)


class MovingAverageEnvelopes(BaseIndicator):
    """
    Moving Average Envelopes
    
    Upper and lower envelopes around a moving average, calculated as percentage bands.
    
    Formula: 
    Upper = MA * (1 + percentage/100)
    Lower = MA * (1 - percentage/100)
    """
    
    def __init__(self):
        super().__init__("MA Envelopes")
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_sma(data: np.ndarray, period: int) -> np.ndarray:
        """Calculate SMA"""
        n = len(data)
        result = np.full(n, np.nan)
        
        for i in range(period - 1, n):
            result[i] = np.mean(data[i - period + 1:i + 1])
        
        return result
    
    @staticmethod
    @jit(nopython=True)
    def _calculate_ema(data: np.ndarray, period: int) -> np.ndarray:
        """Calculate EMA"""
        n = len(data)
        result = np.empty(n)
        alpha = 2.0 / (period + 1)
        
        result[0] = data[0]
        for i in range(1, n):
            result[i] = alpha * data[i] + (1 - alpha) * result[i - 1]
        
        return result
    
    def calculate(self, data: Union[np.ndarray, pd.Series, list],
                 period: int = 20, percentage: float = 2.5,
                 ma_type: str = "SMA") -> Union[Tuple[np.ndarray, np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series, pd.Series]]:
        """
        Calculate Moving Average Envelopes
        
        Parameters:
        -----------
        data : Union[np.ndarray, pd.Series, list]
            Price data (typically closing prices)
        period : int, default=20
            Period for moving average calculation
        percentage : float, default=2.5
            Percentage for envelope calculation
        ma_type : str, default="SMA"
            Type of moving average ("SMA" or "EMA")
            
        Returns:
        --------
        Union[Tuple[np.ndarray, np.ndarray, np.ndarray], Tuple[pd.Series, pd.Series, pd.Series]]
            (upper_envelope, middle_line, lower_envelope) in the same format as input
        """
        validated_data, input_type, index = self.validate_input(data)
        self.validate_period(period, len(validated_data))
        
        # Calculate moving average
        if ma_type.upper() == "SMA":
            ma = self._calculate_sma(validated_data, period)
        elif ma_type.upper() == "EMA":
            ma = self._calculate_ema(validated_data, period)
        else:
            raise ValueError(f"Unsupported MA type: {ma_type}")
        
        # Calculate envelopes
        multiplier = percentage / 100
        upper_envelope = ma * (1 + multiplier)
        lower_envelope = ma * (1 - multiplier)
        
        results = (upper_envelope, ma, lower_envelope)
        return self.format_multiple_outputs(results, input_type, index)